/**
 * 
 */
package de.ibh.StrongRoom.Client.Config;

import java.io.IOException;
import java.util.Enumeration;
import java.util.InvalidPropertiesFormatException;
import java.util.LinkedList;
import java.util.List;

import de.ibh.StrongRoom.Crypto.CryptoBackend.ECryptoPGP_KeyType;

public class SigningConf extends AbstractConf {
    private static final String SRP_SIGNING_SERIAL = "StrongRoom.Signing.Serial";
    private static final String SRP_SIGNING_PUBKEY = "StrongRoom.Signing.PubKey";
    private static final String SRP_SIGNING_SIGUKEY = "StrongRoom.Signing.SigUKey";

    static private SigningConf signingconf = null;
    
    static public SigningConf getInstance() throws InvalidPropertiesFormatException, IOException {
        if(signingconf != null)
        	return signingconf;
        
        return signingconf = new SigningConf();
    }
    
    static public SigningConf getInstanceSilent() {
        if(signingconf != null)
        	return signingconf;
        
        return signingconf = new SigningConf(true);
    }

    private SigningConf() throws InvalidPropertiesFormatException, IOException {
		super();
	}

    private SigningConf(boolean silent) {
		super(silent);
	}

	@Override
	public String getConfName() {
		return "signing";
	}

	
	
    public long getSerial() {
        return Integer.parseInt(prop.getProperty(SRP_SIGNING_SERIAL, "0")); 
    }
    
    public void setSerial(final long serial) {
        prop.setProperty(SRP_SIGNING_SERIAL, Long.toString(serial)); 
    }
	
	
    public String getPubKey() {
        return prop.getProperty(SRP_SIGNING_PUBKEY, ""); 
    }
    
    public void setPubKey(final String pubKey) {
        prop.setProperty(SRP_SIGNING_PUBKEY, pubKey); 
    }
	
	
    public String getSignUKey() {
        return prop.getProperty(SRP_SIGNING_SIGUKEY, ""); 
    }
    
    public void setSignUKey(final String signMKey) {
        prop.setProperty(SRP_SIGNING_SIGUKEY, signMKey); 
    }
	
	
	
	@Override
	public ECryptoPGP_KeyType getSignKeyType() {
		return ECryptoPGP_KeyType.USER_KEY;
	}

	@Override
	public String getSignString() {
		List<String> l = new LinkedList<String>();
		
		for(Enumeration<Object> keys = prop.keys(); keys.hasMoreElements(); ) {
			String key = keys.nextElement().toString();
			
			if(!key.equals(SRP_SIGNING_SIGUKEY))
				l.add(key + "=" + prop.getProperty(key));
		}
		
		return buildSignString(l);
	}
}