/**
 * 
 */
package de.ibh.StrongRoom.Client.SWT;

import java.io.File;
import java.io.IOException;
import java.util.InvalidPropertiesFormatException;
import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.PatternLayout;
import org.apache.lucene.LucenePackage;
import org.bouncycastle.openpgp.PGPPublicKey;

import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Sash;
import org.eclipse.swt.widgets.Text;

import de.ibh.RandomArt.RAImage;
import de.ibh.StrongRoom.Client.Crypto;
import de.ibh.StrongRoom.Client.DB;
import de.ibh.StrongRoom.Client.SRVersion;
import de.ibh.StrongRoom.Client.Config.BaseConf;
import de.ibh.StrongRoom.Client.Config.SigningConf;
import de.ibh.StrongRoom.Client.Config.WindowConf;
import de.ibh.StrongRoom.Crypto.CryptoBackend.ECryptoPGP_KeyType;
import de.ibh.StrongRoom.Crypto.JCEPolicyTest;
import de.ibh.StrongRoom.DB.DBConn;
import de.ibh.StrongRoom.DB.LuceneStore;
import de.ibh.StrongRoom.Exception.SRGenericException;
import de.ibh.StrongRoom.Log.DebugMessage;
import de.ibh.StrongRoom.Log.ErrorMessage.EMessageTypes;
import de.ibh.StrongRoom.l10n.StrongRoomL10N;

/**
 * @author ellguth
 *
 */

public class GUI {

	// enabled Menu & Toolbar icon groups
	static public final int M_DISCONNECTED 	= MainMenu.ITEM_CONNECT | MainMenu.ITEM_PREFERENCES | MainMenu.ITEM_IMPORT;
	static public final int T_DISCONNECTED 	= Toolbar.ITEM_CONNECT;
	
	static public final int M_CONNECTED 	= MainMenu.ITEM_DISCONNECT | MainMenu.ITEM_SINDEX | MainMenu.ITEM_EDIT;
	static public final int T_CONNECTED 	= Toolbar.ITEM_DISCONNECT;
	
	static public final int M_CUSTOMERLIST 	= MainMenu.ITEM_CONNECT;
	static public final int T_CUSTOMERLIST 	= Toolbar.ITEM_CONNECT;
	
	static public final int M_EDITENTRY 	= 0;
	static public final int T_STYLEELEMENTS	= Toolbar.ITEM_BOLD | Toolbar.ITEM_COLOR | Toolbar.ITEM_ITALIC | Toolbar.ITEM_UNDERLINE;
	static public final int COLUMNS_MAX     = 7;
	
	private DebugMessage        logger     = DebugMessage.getInstance();
    private ErrorMsgDialog      errmsg     = null;
	
	private static Shell       	sShell     = null;
	private Text        		searchText;
	private Button              searchNext;
	private Button              searchPrev;
	private Button              searchDetail;
	private Canvas				masterKeyFPRCanvas;
	private Canvas				signingKeyFPRCanvas;
	private Canvas				userKeyFPRCanvas;
	private static MainMenu   	mainmenu   = null;
	private static Toolbar     	toolbar    = null;
    private SearchItem   		di         = null;
	private static TabFolders  	tfolders;
	private static DBConn      	db         = null;
	private BaseConf    		baseconf   = null;
	private SigningConf    		signingconf = null;
	private Crypto				crypto = null;
	private final StrongRoomL10N      l10n = StrongRoomL10N.getInstance(); 

	private Map<String, String> uidMap = null;
	private Listener searchDetailListener;
	
	private GUI() {
		if(JCEPolicyTest.getMaxBits() != 192) {
            MessageBox mb = new MessageBox(new Shell(), SWT.OK | SWT.ICON_ERROR);
            mb.setMessage(l10n._("FATAL ERROR: Your Java environment does not support long crypto keys!"));
            mb.open();
            System.exit(1);
		}
		
        if(!BaseConf.confFileExists()) {
            MessageBox mb = new MessageBox(new Shell(), SWT.OK | SWT.ICON_INFORMATION);
            mb.setMessage(l10n._("You do not have a StrongRoom configuration, the configuration program will be launched to configure StrongRoom."));
            mb.open();
            
            new GUIConfig();
        	
        	BaseConf.dispose();
        }

        Display display = Display.getDefault();
		createSShell();
		
		sShell.open();
		dbinit();

		final int ctTimeOut = 900000;
		final Runnable runnable = new Runnable() {
			public void run() {
				tfolders.tryToCloseAllTabs();
	        }
	    };
		    
		while (!sShell.isDisposed()) {
			if (!display.readAndDispatch()) {
				display.timerExec(ctTimeOut, runnable);
				display.sleep();
				display.readAndDispatch();
	            display.timerExec(-1, runnable);
			}
		}
	}
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// Setup logging
	    BasicConfigurator.configure();

	    SRClassLoader.addJarToClasspath( new File(SRClassLoader.getArchFilename("swt")) );
	    
		new GUI();
		System.exit(0);
	}
	
	public static Image getIcon(final Display display, final String icon) {
		try {
			return new Image(display, GUI.class.getResourceAsStream("/icons/" + icon));
		} catch (Exception e) {
			return display.getSystemImage(SWT.ICON_ERROR);
		}
	}
    
    private void dbinit() {
        mainmenu.enableItems(M_DISCONNECTED);
        toolbar.enableItems(T_DISCONNECTED);
        mainmenu.disableItems(M_CONNECTED);          
        toolbar.disableItems(T_CONNECTED);

        try {
        	crypto = new Crypto(baseconf, true, PassphraseDialog.get()); 
        	
			if((db = DB.getDBConn(crypto, baseconf, signingconf, new InputDialog(sShell), PassphraseDialog.get(), SigningKeyVerificationDialog.get(), sShell, errmsg)) != null) {
	            di.fillTree(db);
	            mainmenu.disableItems(M_DISCONNECTED);
	            toolbar.disableItems(T_DISCONNECTED);
	            mainmenu.enableItems(M_CONNECTED);          
	            toolbar.enableItems(T_CONNECTED);
	            
	            uidMap = crypto.getUIDMap(db);
	        }

			updateRAImages();
		} catch (SRGenericException e) {
			e.printStackTrace();
			
			crypto = null;
			errmsg.printErrorMessage(e.getMessage(), EMessageTypes.TYPE_ERROR);
		}
    }
    
	
    @SuppressWarnings("unchecked")
	private void updateRAImages() {
		PGPPublicKey pk = crypto.retrievePublicSigningKey(ECryptoPGP_KeyType.MASTER_KEY);
		String t = "Master Key";
		if(pk != null) {
			t += ":\n\n";

			Iterator<String> it = pk.getUserIDs();
		    if(it.hasNext())
			    t += it.next() + "\n";

		    t += Crypto.formatFPR(pk.getFingerprint());
		}
		masterKeyFPRCanvas.setToolTipText(t);
		masterKeyFPRCanvas.redraw();
		
		pk = crypto.retrievePublicSigningKey(ECryptoPGP_KeyType.SIGNING_KEY);
		t = "Signing Key";
		if(pk != null) {
			t += ":\n\n";

			Iterator<String> it = pk.getUserIDs();
		    if(it.hasNext())
			    t += it.next() + "\n";

		    t += Crypto.formatFPR(pk.getFingerprint());
		}
		signingKeyFPRCanvas.setToolTipText(t);
    	signingKeyFPRCanvas.redraw();
    	
		pk = crypto.retrievePublicSigningKey(ECryptoPGP_KeyType.USER_KEY);
		t = "User Key";
		if(pk != null) {
			t += ":\n\n";

			Iterator<String> it = pk.getUserIDs();
		    if(it.hasNext())
			    t += it.next() + "\n";

		    t += Crypto.formatFPR(pk.getFingerprint());
		}
		userKeyFPRCanvas.setToolTipText(t);
    	userKeyFPRCanvas.redraw();
    }
	
    protected Listener getSearchDetailListener() {
    	return searchDetailListener;
    }
    
	/**
	 * This method initializes sShell
	 */
	private void createSShell() {
		sShell = new Shell();
		sShell.setText("StrongRoom " + SRVersion.VERSION_STRING);
		sShell.setImage(GUI.getIcon(sShell.getDisplay(), "icon.png"));

		WindowConf windowconf = WindowConf.getInstance();

        logger.debugEnable(windowconf.getDebug());
        errmsg = ErrorMsgDialog.getInstance(sShell);		
        
        try {
			baseconf = BaseConf.getInstance();
		} catch (InvalidPropertiesFormatException e) {
			 errmsg.printErrorMessage(l10n._("Could not parse base configuration: ") + e.getMessage(), EMessageTypes.TYPE_ERROR);
			 closeGUI();
		} catch (IOException e) {
			 errmsg.printErrorMessage(l10n._("Could not read base configuration: ") + e.getMessage(), EMessageTypes.TYPE_ERROR);
			 closeGUI();
		}
		
        try {
			signingconf = SigningConf.getInstance();
		} catch (InvalidPropertiesFormatException e) {
			 errmsg.printErrorMessage(l10n._("Could not parse signing configuration: ") + e.getMessage(), EMessageTypes.TYPE_WARN);
			 signingconf = SigningConf.getInstanceSilent();
		} catch (IOException e) {
			 errmsg.printErrorMessage(l10n._("Could not read signing configuration: ") + e.getMessage(), EMessageTypes.TYPE_WARN);
			 signingconf = SigningConf.getInstanceSilent();
		}
		
        // Restore the size and position of the window.
		int bounds[] = windowconf.getWindowBounds();
        sShell.setBounds(bounds[0], bounds[1], bounds[2], bounds[3]);
		sShell.setLayout(new GridLayout(GUI.COLUMNS_MAX, false));
		
		// Setup the main menu.
		mainmenu = MainMenu.getInstance(sShell);
		mainmenu.disableAllItems();
		mainmenu.enableItems(MainMenu.ITEM_FILE | MainMenu.ITEM_EXIT);
		mainmenu.enableItems(MainMenu.ITEM_HELP | MainMenu.ITEM_ABOUT);
	    mainmenu.enableItems(MainMenu.ITEM_EDIT | MainMenu.ITEM_COPY | MainMenu.ITEM_CUT | MainMenu.ITEM_PASTE);
		mainmenu.enableItems(M_DISCONNECTED);
		mainmenu.setItemSelection(MainMenu.ITEM_COLLAPSELIST, windowconf.getCollapseList());
		mainmenu.setItemSelection(MainMenu.ITEM_HIGHLIGHTALL, windowconf.getHighLightAll());
	    		
		// Setup the toolbar.
	    // Setup Row 1
		toolbar = Toolbar.getInstance(sShell); 
		toolbar.disableAllItems();
		
		// Add a search input field.
		// Setup Row 2
		Label searchLabel = new Label(sShell, SWT.NONE);
		searchLabel.setText(l10n._("Find:"));
		searchLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false, 1, 1));
		
		searchText = new Text(sShell, SWT.BORDER);
		searchText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));
		final Color bgColorTextItem = searchText.getBackground();
		final Color bgColorHighlight = new Color(Display.getDefault(), 255, 100, 100);
		
		searchText.addListener(SWT.Modify, new Listener() {
		        public void handleEvent(Event event) {
		            searchText.setBackground(bgColorTextItem);
		            if(di.selectItemByName(searchText.getText()) == false)
		                searchText.setBackground(bgColorHighlight);
		        }
		    }
		);
		searchText.addListener(SWT.KeyDown, new Listener() {
                public void handleEvent(Event event) {
                    if(event.keyCode == SWT.CR) {
                        di.selectNextItemByName(searchText.getText());
                    }
                }
            }
        );
		
		searchPrev = new Button(sShell, SWT.NONE);
        searchPrev.setToolTipText(l10n._("goto previous match"));
        searchPrev.setImage(getIcon(sShell.getDisplay(), "back.png"));
        searchPrev.setLayoutData(new GridData(SWT.END, SWT.CENTER, false, false, 1, 1));
        searchPrev.addListener(SWT.Selection, new Listener() {
                public void handleEvent(Event event) {
                    di.selectPrevItemByName(searchText.getText());
                }
            }
        );
        
        searchNext = new Button(sShell, SWT.NONE);
        searchNext.setToolTipText(l10n._("goto next match"));
        searchNext.setImage(getIcon(sShell.getDisplay(), "forward.png"));
        searchNext.setLayoutData(new GridData(SWT.END, SWT.CENTER, false, false, 1, 1));
        searchNext.addListener(SWT.Selection, new Listener() {
                public void handleEvent(Event event) {
                    di.selectNextItemByName(searchText.getText());
                }
            }
        );
        
        searchDetail = new Button(sShell, SWT.NONE);
        searchDetail.setToolTipText(l10n._("Content Search"));
        searchDetail.setImage(getIcon(sShell.getDisplay(), "search.png"));
        searchDetail.setLayoutData(new GridData(SWT.END, SWT.CENTER, false, false, 1, 1));
        searchDetailListener = new Listener() {
            public void handleEvent(Event event) {
                if(!searchDetail.getEnabled())
                    return;
                
                try {
                    new SearchDetail(sShell, DBConn.getInstance(), crypto, baseconf);
                } catch (SRGenericException e) {
                    errmsg.printErrorMessage(l10n._("Error creating search window: ") + e.getLocalizedMessage(), EMessageTypes.TYPE_ERROR);
                }
            }
        };
        
        searchDetail.addListener(SWT.Selection, searchDetailListener);
        mainmenu.addItemListener(MainMenu.ITEM_FIND, searchDetailListener);
        
        Group fprGroup = new Group(sShell, SWT.NONE);
        fprGroup.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, false, false, 1, 1));
        fprGroup.setLayout(new GridLayout(3, true));
        
        GridData gdcanvas = new GridData(SWT.CENTER, SWT.CENTER, false, false, 1, 1);
        gdcanvas.minimumHeight = RAImage.getDefaultHeight();
        gdcanvas.minimumWidth = RAImage.getDefaultWidth();
        gdcanvas.heightHint = RAImage.getDefaultHeight();
        gdcanvas.widthHint = RAImage.getDefaultWidth();
        
        masterKeyFPRCanvas = new Canvas(fprGroup, SWT.NONE);
        masterKeyFPRCanvas.addPaintListener(new PaintListener() {
            public void paintControl(PaintEvent e) {
            	String fpr = null;

            	if(crypto != null) {
	            	PGPPublicKey pk = crypto.retrievePublicSigningKey(ECryptoPGP_KeyType.MASTER_KEY);
	            	if(pk != null)
	            		fpr = Crypto.formatFPR(pk.getFingerprint());
            	}

            	final Image img = RAImage.create(fpr, sShell.getDisplay());
                e.gc.drawImage(img, 0, 0);
            }
        });
        masterKeyFPRCanvas.setLayoutData(gdcanvas);
        masterKeyFPRCanvas.setToolTipText("Master Key");
        
        signingKeyFPRCanvas = new Canvas(fprGroup, SWT.NONE);
        signingKeyFPRCanvas.addPaintListener(new PaintListener() {
            public void paintControl(PaintEvent e) {
            	String fpr = null;

            	if(crypto != null) {
	            	PGPPublicKey pk = crypto.retrievePublicSigningKey(ECryptoPGP_KeyType.SIGNING_KEY);
	            	if(pk != null)
	            		fpr = Crypto.formatFPR(pk.getFingerprint());
            	}

            	final Image img = RAImage.create(fpr, sShell.getDisplay());
                e.gc.drawImage(img, 0, 0);
            }
        });
        signingKeyFPRCanvas.setLayoutData(gdcanvas);        
        signingKeyFPRCanvas.setToolTipText("Signing Key");
        
        userKeyFPRCanvas = new Canvas(fprGroup, SWT.NONE);
        userKeyFPRCanvas.addPaintListener(new PaintListener() {
            public void paintControl(PaintEvent e) {
            	String fpr = null;

            	if(crypto != null) {
	            	PGPPublicKey pk = crypto.retrievePublicSigningKey(ECryptoPGP_KeyType.USER_KEY);
	            	if(pk != null)
	            		fpr = Crypto.formatFPR(pk.getFingerprint());        		
            	}

            	final Image img = RAImage.create(fpr, sShell.getDisplay());
                e.gc.drawImage(img, 0, 0);
            }
        });
        userKeyFPRCanvas.setLayoutData(gdcanvas);        
        userKeyFPRCanvas.setToolTipText("User Key");
		
        // Setup Row 3 
	    tfolders = new TabFolders(sShell, this, SWT.BORDER);
	    di       = new SearchItem(tfolders.getParent(), this, tfolders, errmsg);
      
        tfolders.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, GUI.COLUMNS_MAX, 1));
    
        Listener closeListener = new Listener() {
            public void handleEvent(Event event) {
                if(tfolders.tryToCloseAllTabs()) {
                    if(event.widget instanceof Shell)
                        event.doit = true;
                    else
                        GUI.closeGUI();
                }
                else
                    event.doit = false;
            }
        };
		toolbar.addItemListener(Toolbar.ITEM_EXIT, closeListener);
		mainmenu.addItemListener(MainMenu.ITEM_EXIT, closeListener);
		sShell.addListener(SWT.Close, closeListener);
		
		toolbar.addItemListener(Toolbar.ITEM_CONNECT, new Listener() {
	        public void handleEvent(Event event) {
	        	db = null;
	        	dbinit();
	        }
	    }
		);

		toolbar.addItemListener(Toolbar.ITEM_DISCONNECT, new Listener() {
	        public void handleEvent(Event event) {
	            if(tfolders.tryToCloseAllTabs()) {
	                closeDB(db);
		        
    	            di.clearTree();
    		        mainmenu.disableItems(M_CONNECTED);
    		        mainmenu.enableItems(M_DISCONNECTED);
    		        toolbar.disableItems(T_CONNECTED);
    		        toolbar.enableItems(T_DISCONNECTED);
	            }
	        }
	    });
		
		
        mainmenu.addItemListener(MainMenu.ITEM_ABOUT, new Listener() {
            public void handleEvent(Event event) {

            	String about =
            			"StrongRoom " + SRVersion.VERSION_STRING + "\n\n" +
            			"2009 - 2015 (C) IBH IT-Service GmbH\n\n" + 
            			l10n._("Includes") +  ":\n" +
            				Crypto.getInfo() + "\n" +
            				LucenePackage.get().getImplementationTitle() + " " + LucenePackage.get().getImplementationVersion();
                errmsg.printErrorMessage(about, EMessageTypes.TYPE_INFO);
            }
        });
        

		mainmenu.addItemListener(MainMenu.ITEM_CONNECT, new Listener() {
	        public void handleEvent(Event event) {
	        	db = null;
	        	dbinit();
	        }
	    });
		
		mainmenu.addItemListener(MainMenu.ITEM_DISCONNECT, new Listener() {
	        public void handleEvent(Event event) {
	            if(tfolders.tryToCloseAllTabs()) {
    	            closeDB(db);
    		        
    	            di.clearTree();
                    mainmenu.disableItems(M_CONNECTED);
                    mainmenu.enableItems(M_DISCONNECTED);
                    toolbar.disableItems(T_CONNECTED);
                    toolbar.enableItems(T_DISCONNECTED);
	            }
	        }
	    });
		
		mainmenu.addItemListener(MainMenu.ITEM_IMPORT, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
                new ImportData();
            }
        });
		
		mainmenu.addItemListener(MainMenu.ITEM_PREFERENCES, new Listener() {
	        public void handleEvent(Event event) {
	        	new GUIConfig();
	        	try {
					baseconf.reload();
				} catch (Exception e) {
		            MessageBox mb = new MessageBox(new Shell(), SWT.OK | SWT.ICON_ERROR);
		            mb.setMessage(l10n._("There was an error reloading you configuration: ") + e.getLocalizedMessage());
		            mb.open();
		            
		            System.exit(1);
				}
	        }
	    });
		
		mainmenu.addItemListener(MainMenu.ITEM_SINDEX, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
                try {
					new SearchIndexer(sShell, DBConn.getInstance(), baseconf, crypto);
				} catch (SRGenericException e) {
					errmsg.printErrorMessage(l10n._("Error building search index: ") + e.getLocalizedMessage(), EMessageTypes.TYPE_ERROR);
				}
            }
        });
		
		toolbar.enableItems(Toolbar.ITEM_EXIT);
		toolbar.enableItems(T_DISCONNECTED);
		
		searchText.setFocus();

		if(windowconf.getDebug()) {
		    // Setup the slider for resize the tree and the logging area..
	        // Setup Row 4
			final Sash sash = new Sash(sShell, SWT.HORIZONTAL);
			sash.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false, GUI.COLUMNS_MAX, 1));
			
			// Setup the logging area.
	        // Setup Row 5
			StyledText logText = new StyledText(sShell, SWT.BORDER);
			logText.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false, GUI.COLUMNS_MAX, 1));
			
			StyledTextAppender tap = new StyledTextAppender();
			tap.setLayout(new PatternLayout("%d [%t] %-5p %m%n"));
			tap.setStyledText(logText);
			tap.getStyledText().setEditable(false);
			logger.addAppender(tap);
		    
	        sash.addListener(SWT.Selection, new Listener() {
	            public void handleEvent(Event e) {
	              sash.setBounds(e.x, e.y, e.width, e.height);
	              sShell.layout();
	            }
	        });
		}
		
	    logger.info("StrongRoom " + SRVersion.VERSION_STRING);
        
        sShell.setTabList(new Control[]{searchText, tfolders.getParent()});
	}
	
	
	public Crypto getCrypto() {
	    return crypto;
	}

	public String getUIDbyFPR(final String fpr) {
		if(uidMap.containsKey(fpr))
			return uidMap.get(fpr);

		return "?";
	}
	
	public void setSearchTextEnabled(boolean active) {
	    searchText.setEnabled(active);
	    searchNext.setEnabled(active);
	    searchPrev.setEnabled(active);
	    searchDetail.setEnabled(active);
	}
	

	public static void closeDB(DBConn db) {
		db.disconnect();
	}
	
	public static TabFolders getTfolder(){
		return tfolders;
	}
	
	
	public static void closeGUI() {
		if(mainmenu != null) {
			/* Save the window dimensions before quit. */
			WindowConf windowconf = WindowConf.getInstance();
		
			Rectangle bounds = sShell.getBounds();
		
			windowconf.setWindowBounds(new int[] {bounds.x, bounds.y, bounds.width, bounds.height});
			windowconf.setCollapseList(mainmenu.getItemSelection(MainMenu.ITEM_COLLAPSELIST));
			windowconf.setHighlightAll(mainmenu.getItemSelection(MainMenu.ITEM_HIGHLIGHTALL));
			windowconf.saveConfigSilent();
		}
		
		sShell.close();
	    System.exit(0);
	}


	public static String escapeStr(String s) {
		if(s == null)
			return "";

		return s.replaceAll("&", "&&");
	}


	public static String shrinkStr(String s, final int min, final int max) {
		if(s == null)
			return "";

		if(s.length() <= min)
			return s;
		
		int i = Math.max(
			Math.min(Math.max(min, s.indexOf(' ', min) - 1), max),
			Math.max(
					Math.max(
							Math.min(Math.max(min, s.indexOf('-', min) - 1), max),
							Math.min(Math.max(min, s.indexOf('&', min) - 1), max)
					),
					Math.min(Math.max(min, s.indexOf('.', min) - 1), max)
			)
		);

		return s.substring(0, i + 1).trim();
	}
}
