/**
 * 
 */
package de.ibh.StrongRoom.Client.SWT;

import java.io.File;
import java.sql.Driver;
import java.sql.DriverManager;
import java.util.Enumeration;
import java.util.Iterator;
import org.apache.log4j.BasicConfigurator;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.media_art_online.gnupg.GnuPG;
import org.media_art_online.gnupg.GnuPGUser;

import de.ibh.RandomArt.RAImage;
import de.ibh.StrongRoom.Client.Crypto;
import de.ibh.StrongRoom.Client.Misc;
import de.ibh.StrongRoom.Client.Config.AutoConf;
import de.ibh.StrongRoom.Client.Config.BaseConf;
import de.ibh.StrongRoom.Client.Config.BaseConf.EDBCredentialMethod;
import de.ibh.StrongRoom.Client.Config.BaseConf.ESecretKeySource;
import de.ibh.StrongRoom.Crypto.CryptoBackend.ECryptoPGP_KeyType;
import de.ibh.StrongRoom.DB.DBConn;
import de.ibh.StrongRoom.Exception.SRGenericCryptoEngineException;
import de.ibh.StrongRoom.Exception.SRGenericCryptoModelException;
import de.ibh.StrongRoom.Exception.SRGenericException;
import de.ibh.StrongRoom.Log.DebugMessage;
import de.ibh.StrongRoom.Log.ErrorMessage.EMessageTypes;
import de.ibh.StrongRoom.l10n.StrongRoomL10N;

/**
 * @author liske
 *
 */

public class GUIConfig {
	private Shell sShell;
	private Color col_valid;
	private Color col_invalid;

	private BaseConf conf;
	private Crypto crypto = null;
	private DebugMessage logger = DebugMessage.getInstance();

	private StrongRoomL10N l10n = StrongRoomL10N.getInstance();

	private Combo dbJDBCDriverCombo;
	private Text dbURLText;
	private Combo dbCredentialMethodCombo;
	private Text dbUsernameText;
	private Text dbPasswordText;

	private Text pgpSecretKeyFPRText;
	private Button pgpSecretKeyFPRButton;
	private Canvas pgpSecretKeyFPRCanvas;
	private Combo pgpSecretKeyMethodCombo;
	private Text pgpSecretKeyGnuPGText;
	private Text pgpSecretKeyDataText;

	private Text pgpMasterKeyFPRText;
	private Canvas pgpMasterKeyFPRCanvas;
	private Text pgpMasterKeyDataText;

	private Label verifyLabel;
	private Text signatureText;
	
	private Button saveButton;
	private boolean changed;

	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		/* setup logging */
	    BasicConfigurator.configure();

	    SRClassLoader.addJarToClasspath( new File(SRClassLoader.getArchFilename("swt")) );

	    new GUIConfig();
		System.exit(0);
	}
	
	public GUIConfig() {
		Display display = Display.getDefault();

		conf = BaseConf.getInstanceSilent();
		
		logger.debugEnable(true);
		createSShell();
		sShell.open();

		while (!sShell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}
	}
	
	public void setConfig() throws SRGenericCryptoModelException {
		conf.setDBJDBC(dbJDBCDriverCombo.getText());
		conf.setDBURL(dbURLText.getText());
		conf.setDBCredentialMethod(EDBCredentialMethod.valueOf(dbCredentialMethodCombo.getText()));
		conf.setDBUsername(dbUsernameText.getText());

		conf.setSecretKeyFPR(pgpSecretKeyFPRText.getText());
		conf.setSecretKeyKeyMethod(ESecretKeySource.valueOf(pgpSecretKeyMethodCombo.getText()));
		conf.setSecretKeyGPGBin(pgpSecretKeyGnuPGText.getText());
		if(pgpSecretKeyMethodCombo.getText().equals(ESecretKeySource.PLAIN.name()))
			conf.setSecretKeyData(pgpSecretKeyDataText.getText());
		else
			conf.setSecretKeyData("");
		
		conf.setMasKeyFPR(pgpMasterKeyFPRText.getText());
		conf.setMasKeyData(pgpMasterKeyDataText.getText());

		if(dbCredentialMethodCombo.getText().equals(EDBCredentialMethod.ENCRYPTED.name())) {
			Crypto crypto = new Crypto(conf, false, PassphraseDialog.get());
			
			String eDb_password;
			try {
				eDb_password = Misc.b2s(crypto.encrypt(
				        new ECryptoPGP_KeyType[] {ECryptoPGP_KeyType.USER_KEY},
				        dbPasswordText.getText().getBytes(),
				        null,
				        null
				));
			} catch (SRGenericCryptoEngineException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				
				return;
			} 
			
			if(eDb_password == null || eDb_password.equals(""))
                throw new SRGenericCryptoModelException(l10n._("Encryption of database password failed!")) {
					private static final long serialVersionUID = 1L;
                };

		    conf.setDBPassword(eDb_password);
		}
		else
			conf.setDBPassword("");
	}
	
	/**
	 * This method initializes sShell
	 */
	private void createSShell() {
		sShell = new Shell();

        sShell.setText(l10n._("StrongRoom Configuration"));
		sShell.setLayout(new GridLayout());
	
		col_valid = sShell.getDisplay().getSystemColor(SWT.COLOR_DARK_GREEN); 
		col_invalid = sShell.getDisplay().getSystemColor(SWT.COLOR_RED);
		
		final ScrolledComposite sc = new ScrolledComposite(sShell, SWT.V_SCROLL);		
		sc.setLayoutData(new GridData(SWT.FILL, SWT.BEGINNING, true, true));
		
		Composite c = new Composite(sc, SWT.NONE);
		c.setLayout(new GridLayout(2, false));

	
		GridData gdmulti = new GridData(SWT.FILL, SWT.FILL, true, false);
		gdmulti.minimumHeight = 80;
		gdmulti.heightHint = 80;

		GridData gdmulti2 = new GridData(SWT.FILL, SWT.FILL, true, false, 2, 1);
		gdmulti2.minimumHeight = 100;
		gdmulti2.heightHint = 100;

		GridData gdmulti3 = new GridData(SWT.FILL, SWT.FILL, true, false, 3, 1);
		gdmulti3.minimumHeight = 100;
		gdmulti3.heightHint = 100;
		
		ModifyListener ml = new ModifyListener() {
			@Override
			public void modifyText(ModifyEvent arg0) {
				if(!changed) {
					changed = true;
					updateVerifyLabel();
				}
			}
		};
		
		/* Intro */

		Label l = new Label(c, SWT.WRAP);
		l.setText(l10n._("Whenever you change any setting of your configuration you have to generate a new signature!"));
		l.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false, 2, 1));

		l = new Label(c, SWT.WRAP);
		l.setText(l10n._("To get an initial configuration you might try to use the 'auto configure' feature."));
		l.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false, 2, 1));
		
		final Button buttonAutoConf = new Button(c, SWT.NONE);
		buttonAutoConf.setText(l10n._("auto configure"));
		buttonAutoConf.setImage(GUI.getIcon(sShell.getDisplay(), "search.png"));
		buttonAutoConf.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false, 2, 1));
		
		
		/* DB */
	    
		Group g = new Group(c, SWT.SHADOW_ETCHED_IN);
		g.setText(l10n._("Database"));
		g.setLayout(new GridLayout(2, false));
		g.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));

		
		l = new Label(g, SWT.NONE);
		l.setText(l10n._("JDBC Driver:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

		DBConn.loadJDBCClasses();
		dbJDBCDriverCombo = new Combo(g, SWT.DROP_DOWN | SWT.BORDER);
		dbJDBCDriverCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		dbJDBCDriverCombo.setText(conf.getDBCredentialMethod().name());
		for(Enumeration<Driver> en = DriverManager.getDrivers(); en.hasMoreElements(); ) {
			dbJDBCDriverCombo.add(en.nextElement().getClass().getCanonicalName());
		}
		dbJDBCDriverCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		dbJDBCDriverCombo.setText(conf.getDBJDBC());
		dbJDBCDriverCombo.addModifyListener(ml);

		
		l = new Label(g, SWT.NONE);
		l.setText(l10n._("URL:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
		
		dbURLText = new Text(g, SWT.BORDER);
		dbURLText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		dbURLText.setText(conf.getDBURL());
		dbURLText.addModifyListener(ml);
		
		
		l = new Label(g, SWT.NONE);
		l.setText(l10n._("Method:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
		
		dbCredentialMethodCombo = new Combo(g, SWT.DROP_DOWN | SWT.BORDER | SWT.READ_ONLY);
		dbCredentialMethodCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		dbCredentialMethodCombo.setText(conf.getDBCredentialMethod().name());
		EDBCredentialMethod vc[] = EDBCredentialMethod.values();
		for(int i = 0; i < vc.length; i++)
			dbCredentialMethodCombo.add(vc[i].name());
		dbCredentialMethodCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		dbCredentialMethodCombo.select(dbCredentialMethodCombo.indexOf(conf.getDBCredentialMethod().name()));
		dbCredentialMethodCombo.addSelectionListener(new SelectionListener() {
			
			@Override
			public void widgetSelected(SelectionEvent arg0) {
				if(dbUsernameText != null)
					dbUsernameText.setEnabled(dbCredentialMethodCombo.getText().equals(EDBCredentialMethod.ENCRYPTED.name()));

				if(dbPasswordText != null)
					dbPasswordText.setEnabled(dbCredentialMethodCombo.getText().equals(EDBCredentialMethod.ENCRYPTED.name()));
				
				if(!dbCredentialMethodCombo.getText().equals(EDBCredentialMethod.ENCRYPTED.name())) {
					if(dbUsernameText != null)
						dbUsernameText.setText("");

					if(dbPasswordText != null)
						dbPasswordText.setText("");
				}
			}
			
			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}
		});
		dbCredentialMethodCombo.addModifyListener(ml);

		
		l = new Label(g, SWT.NONE);
		l.setText(l10n._("Username:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
		
		dbUsernameText = new Text(g, SWT.BORDER);
		dbUsernameText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		dbUsernameText.setText(conf.getDBUsername());
		dbUsernameText.addModifyListener(ml);
		dbUsernameText.setEnabled(dbCredentialMethodCombo.getText().equals(EDBCredentialMethod.ENCRYPTED.name()));

		
		l = new Label(g, SWT.NONE);
		l.setText(l10n._("Password:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.TOP, false, false));
			
		dbPasswordText = new Text(g, SWT.BORDER);
		dbPasswordText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		dbPasswordText.setEchoChar('x');
		dbPasswordText.addModifyListener(ml);
		dbPasswordText.setEnabled(dbCredentialMethodCombo.getText().equals(EDBCredentialMethod.ENCRYPTED.name()));

		
		/* SecretKey */
		
		g = new Group(c, SWT.SHADOW_ETCHED_IN);
		g.setText(l10n._("Secret Key"));
		g.setLayout(new GridLayout(4, false));
		g.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));
		

		l = new Label(g, SWT.NONE);
		l.setText(l10n._("Key source:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
			
		pgpSecretKeyMethodCombo = new Combo(g, SWT.DROP_DOWN | SWT.BORDER | SWT.READ_ONLY);
		ESecretKeySource vs[] = ESecretKeySource.values();
		for(int i = 0; i < vs.length; i++)
			pgpSecretKeyMethodCombo.add(vs[i].name());
		pgpSecretKeyMethodCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 3, 1));
		pgpSecretKeyMethodCombo.select(pgpSecretKeyMethodCombo.indexOf(conf.getSecretKeyKeyMethod().name()));
		pgpSecretKeyMethodCombo.addSelectionListener(new SelectionListener() {
			
			@Override
			public void widgetSelected(SelectionEvent arg0) {
				if(pgpSecretKeyFPRButton != null)
					pgpSecretKeyFPRButton.setEnabled(pgpSecretKeyMethodCombo.getText().equals(ESecretKeySource.GNUPG.name()));

				if(pgpSecretKeyGnuPGText != null)
					pgpSecretKeyGnuPGText.setEnabled(pgpSecretKeyMethodCombo.getText().equals(ESecretKeySource.GNUPG.name()));

				if(pgpSecretKeyDataText != null)
					pgpSecretKeyDataText.setEnabled(pgpSecretKeyMethodCombo.getText().equals(ESecretKeySource.PLAIN.name()));
			}
			
			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}
		});
		pgpSecretKeyMethodCombo.addModifyListener(ml);
		

		l = new Label(g, SWT.NONE);
		l.setText(l10n._("Fingerprint:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
			
		pgpSecretKeyFPRText = new Text(g, SWT.BORDER);
		pgpSecretKeyFPRText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		pgpSecretKeyFPRText.setText(conf.getSecretKeyFPR());
		pgpSecretKeyFPRText.addFocusListener(
				new FocusListener() {
					
					@Override
					public void focusLost(FocusEvent arg0) {
						pgpSecretKeyFPRCanvas.redraw();
					}
					
					@Override
					public void focusGained(FocusEvent arg0) {
					}
				}
		);
		pgpSecretKeyFPRText.addModifyListener(ml);

		pgpSecretKeyFPRButton = new Button(g, SWT.BORDER);
		pgpSecretKeyFPRButton.setText(l10n._("Select"));
		pgpSecretKeyFPRButton.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, false, false));
		pgpSecretKeyFPRButton.setEnabled(pgpSecretKeyMethodCombo.getText().equals(ESecretKeySource.GNUPG.name()));
		pgpSecretKeyFPRButton.addListener(SWT.Selection, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
            	final Display dsp = arg0.display;
                final Shell s = new Shell(sShell);
                s.setText(l10n._("Select your GPG key"));
                s.setLayout(new FillLayout());
                s.addListener(SWT.Traverse, new Listener() {
                	public void handleEvent(Event event) {
                		switch (event.detail) {
                			case SWT.TRAVERSE_ESCAPE:
                				s.close();
                				event.detail = SWT.TRAVERSE_NONE;
                				event.doit = false;
                				break;
                		}
                    }
                });

                final Table tbl = new Table(s, SWT.BORDER);
                tbl.addSelectionListener(new SelectionListener() {
					@Override
					public void widgetSelected(SelectionEvent arg0) {
					}
					
					@Override
					public void widgetDefaultSelected(SelectionEvent arg0) {
						TableItem[] tis = tbl.getSelection();
						if(tis.length == 1) {
							pgpSecretKeyFPRText.setText(tis[0].getData().toString());
							s.dispose();
							pgpSecretKeyFPRCanvas.redraw();
						}
					}
				});
                
    	        GnuPG.setPath(pgpSecretKeyGnuPGText.getText());
    	        GnuPG.updateKeys();
    	        new GnuPG();
    	        boolean found = false;

    	        for(Iterator<GnuPGUser> it = GnuPG.listSecretKeys().iterator(); it.hasNext(); ) {
    	        	GnuPGUser u = it.next();
    	        	
    	        	TableItem ti = new TableItem(tbl, SWT.NONE);
    	        	
                    ti.setImage(dsp.getSystemImage(SWT.ICON_INFORMATION));
                    final String fpr = Crypto.formatFPR(u.getPrimaryKey().getFingerprint());
                    ti.setImage(RAImage.create(fpr, dsp));
                    ti.setText(u.getIdentifyingName().toString() + " [" + fpr + "]");
                    ti.setData(fpr);
                    
                    found = true;
    	        }
                
    	        if(!found) {
    	        	s.dispose();

    	        	final MessageBox mb = new MessageBox(sShell);
    	        	mb.setText(l10n._("gpg error"));
    	        	mb.setMessage(l10n._("Sorry, gpg does not return any secret keys!"));
    	        	mb.open();

    	        	return;
    	        }
    	        
                s.open();
                while(!s.isDisposed())
                	if(!dsp.readAndDispatch())
                		dsp.sleep();
                s.dispose();
            }
        });
		
		pgpSecretKeyFPRCanvas = new Canvas(g, SWT.NONE);
		pgpSecretKeyFPRCanvas.addPaintListener(new PaintListener() {
			public void paintControl(PaintEvent e) {
				String s = pgpSecretKeyFPRText.getText();

				if("".equals(s))
					s = null;

				final Image img = RAImage.create(s, sShell.getDisplay());
				e.gc.drawImage(img, 0, 0);
			}
		});
		pgpSecretKeyFPRCanvas.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, false, false));


		l = new Label(g, SWT.NONE);
		l.setText(l10n._("GnuPG binary:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
			
		pgpSecretKeyGnuPGText = new Text(g, SWT.BORDER);
		pgpSecretKeyGnuPGText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 3, 1));
		pgpSecretKeyGnuPGText.setText(conf.getSecretKeyGPGBin());
		pgpSecretKeyGnuPGText.setEnabled(pgpSecretKeyMethodCombo.getText().equals(ESecretKeySource.GNUPG.name()));
		pgpSecretKeyGnuPGText.addModifyListener(ml);


		l = new Label(g, SWT.NONE);
		l.setText(l10n._("Secret Key:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.TOP, false, false));
			
		pgpSecretKeyDataText = new Text(g, SWT.BORDER | SWT.MULTI | SWT.V_SCROLL);
		pgpSecretKeyDataText.setLayoutData(gdmulti3);
		pgpSecretKeyDataText.setText(conf.getSecretKeyData());
		pgpSecretKeyDataText.setEnabled(pgpSecretKeyMethodCombo.getText().equals(ESecretKeySource.PLAIN.name()));
		pgpSecretKeyDataText.addModifyListener(ml);
		
		
		
		/* MasterKey */
		
		g = new Group(c, SWT.SHADOW_ETCHED_IN);
		g.setText(l10n._("Master Key"));
		g.setLayout(new GridLayout(3, false));
		g.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));
		

		l = new Label(g, SWT.NONE);
		l.setText(l10n._("Fingerprint:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
			
		pgpMasterKeyFPRText = new Text(g, SWT.BORDER);
		pgpMasterKeyFPRText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		pgpMasterKeyFPRText.setText(conf.getMasKeyFPR());
		pgpMasterKeyFPRText.addFocusListener(
				new FocusListener() {
					
					@Override
					public void focusLost(FocusEvent arg0) {
						pgpMasterKeyFPRCanvas.redraw();
					}
					
					@Override
					public void focusGained(FocusEvent arg0) {
					}
				}
		);
		pgpMasterKeyFPRText.addModifyListener(ml);

		pgpMasterKeyFPRCanvas = new Canvas(g, SWT.NONE);
		pgpMasterKeyFPRCanvas.addPaintListener(new PaintListener() {
			public void paintControl(PaintEvent e) {
				String s = pgpMasterKeyFPRText.getText();
				
				if("".equals(s))
					s = null;

				final Image img = RAImage.create(s, sShell.getDisplay());
				e.gc.drawImage(img, 0, 0);
			}
		});
		pgpMasterKeyFPRCanvas.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, false, false));


		l = new Label(g, SWT.NONE);
		l.setText(l10n._("Public key:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.TOP, false, false));
			
		pgpMasterKeyDataText = new Text(g, SWT.BORDER | SWT.MULTI | SWT.V_SCROLL);
		pgpMasterKeyDataText.setLayoutData(gdmulti2);
		pgpMasterKeyDataText.setText(conf.getMasKeyData());
		pgpMasterKeyDataText.addModifyListener(ml);

		
		
		/* Signature */
		
		g = new Group(c, SWT.SHADOW_ETCHED_IN);
		g.setText(l10n._("Signature"));
		g.setLayout(new GridLayout(2, false));
		g.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));
		
		
		Button genSigButton = new Button(g, SWT.BORDER);
		genSigButton.setText(l10n._("Generate"));
		genSigButton.setLayoutData(new GridData(SWT.BEGINNING, SWT.TOP, false, false));
		
		signatureText = new Text(g, SWT.BORDER | SWT.MULTI | SWT.V_SCROLL);
		signatureText.setLayoutData(gdmulti);
		signatureText.setEditable(false);
		signatureText.setText(conf.getSignature());
		
		genSigButton.addListener(SWT.Selection, new Listener() {
            @Override
            public void handleEvent(Event arg0) {
            	signatureText.setText("");

				try {
					setConfig();

					crypto = new Crypto(conf, false, PassphraseDialog.get());
				    
				    byte [] newSig = crypto.createSignature(ECryptoPGP_KeyType.USER_KEY, conf.getSignString().getBytes(), null, sShell);
				    
	                if(newSig != null) {
	                	String sig = Misc.b2s(newSig);
	                    signatureText.setText(sig);
	                    conf.setSignature(sig);
	                    changed = false;
	                }
				} catch (SRGenericException e) {
					final MessageBox mb = new MessageBox(sShell);
					mb.setText(l10n._("Signing failed"));
					mb.setMessage(l10n._("Could not create signature: ") + e.getLocalizedMessage() + "\n" + e.getClass().getSimpleName());
					mb.open();
                } finally {
                    updateVerifyLabel();
                }
            }
        });


		l = new Label(g, SWT.NONE);
		l.setText(l10n._("Status:"));
		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

		verifyLabel = new Label(g, SWT.NONE);
		verifyLabel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		verifyLabel.setText(l10n._("not verified."));
		
		
		/* Buttons */
		buttonAutoConf.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetSelected(SelectionEvent arg0) {
				try {
					AutoConf ac = AutoConf.getInstance();
					
					String s = ac.getDBURL(dbJDBCDriverCombo.getText());
					if(s != null && !s.isEmpty())
						dbURLText.setText(s);

					s = ac.getMasKeyFPR();
					if(s != null && !s.isEmpty())
						pgpMasterKeyFPRText.setText(s);
					
					s = ac.getMasKeyData();
					if(s != null && !s.isEmpty())
						pgpMasterKeyDataText.setText(s);
				} catch (Exception e) {
					final ErrorMsgDialog errmsg = ErrorMsgDialog.getInstance(sShell);
					errmsg.printErrorMessage(l10n._("Auto configuration did fail: ") + e.getLocalizedMessage(), EMessageTypes.TYPE_ERROR);
				}
			}
			
			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}
		});
		
		saveButton = new Button(c, SWT.BORDER);
		saveButton.setText(l10n._("Save"));
		saveButton.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
		saveButton.setEnabled(false);

		saveButton.addListener(SWT.Selection, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
                conf.saveConfigSilent();
                sShell.dispose();
            }
        });
		
        Button exitButton = new Button(c, SWT.BORDER);
        exitButton.setText(l10n._("Abort"));
        exitButton.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));		

        exitButton.addListener(SWT.Selection, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
                sShell.dispose();
            }
        });
		
	    sc.setContent(c);
	    sc.setExpandHorizontal(true);
	    sc.setExpandVertical(true);
	    sc.setMinSize(c.computeSize(SWT.DEFAULT, SWT.DEFAULT));

		dbURLText.setFocus();
		
    	changed = false;
		try {
			crypto = new Crypto(conf, true, PassphraseDialog.get());
		} catch (SRGenericCryptoModelException e1) {
        	changed = true;
		}

		if(conf.getDBCredentialMethod() == EDBCredentialMethod.ENCRYPTED && crypto != null) {
			String db_password = null;
			
			if(conf.getDBPassword() != null && !conf.getDBPassword().equals("")) {
			    try {
					db_password = Misc.b2s(crypto.decrypt(ECryptoPGP_KeyType.USER_KEY, conf.getDBPassword().getBytes(), PassphraseDialog.get(), null, sShell));
				} catch (Exception e) {
					db_password = null;

					MessageBox mb = new MessageBox(sShell, SWT.OK | SWT.ICON_ERROR);
e.printStackTrace();
					mb.setMessage(l10n._("Decryption of the database password failed:\n\n") + e.getLocalizedMessage() + l10n._("\n\nYou have to enter the DB password before saving the config, otherwise the DB password will be lost!"));
	                mb.open();
	            	changed = true;
				}
			}

			if(db_password != null) {
				boolean ch = changed;
            	dbPasswordText.setText(db_password);
            	changed = ch;
			}
		}

		updateVerifyLabel();
	}

	private void updateVerifyLabel() {
		if(changed) {
			verifyLabel.setText(l10n._("Configuration has been changed - you have to sign it."));
			verifyLabel.setForeground(col_invalid);

			saveButton.setEnabled(false);
		
			verifyLabel.redraw();
			
			return;
		}

		try {
			crypto = new Crypto(conf, true, PassphraseDialog.get());

			verifyLabel.setText(l10n._("You configuration is valid."));
			verifyLabel.setForeground(col_valid);

			saveButton.setEnabled(true);
		} catch (SRGenericCryptoModelException e) {
			verifyLabel.setText(e.getMessage());
			verifyLabel.setForeground(col_invalid);

			saveButton.setEnabled(false);
		}
		
		verifyLabel.redraw();
	}
}
