package de.ibh.StrongRoom.Client.SWT;

import org.eclipse.swt.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.widgets.*;

import de.ibh.StrongRoom.Client.UserInput;
import de.ibh.StrongRoom.l10n.StrongRoomL10N;

/*
 * http://www.java2s.com/Code/Java/SWT-JFace-Eclipse/Howtocreateyourowndialogclasses.htm
 *
 * Rob Warner (rwarner@interspatial.com)
 * Robert Harris (rbrt_harris@yahoo.com)
 */

public class InputDialog extends Dialog implements UserInput {
	  private String message;
	  private String input;
	  private StrongRoomL10N l10n = StrongRoomL10N.getInstance();
	  
  
	  /**
	   * InputDialog constructor
	   * 
	   * @param parent the parent
	   */
	  public InputDialog(Shell parent) {
	    // Pass the default styles here
	    this(parent, SWT.DIALOG_TRIM | SWT.APPLICATION_MODAL);
	  }

	  /**
	   * InputDialog constructor
	   * 
	   * @param parent the parent
	   * @param style the style
	   */
	  public InputDialog(Shell parent, int style) {
	    // Let users override the default styles
	    super(parent, style);
	    setText(l10n._("Input Dialog"));
	    setMessage(l10n._("Please enter a value:"));
	  }

	  /**
	   * Gets the message
	   * 
	   * @return String
	   */
	  public String getMessage() {
	    return message;
	  }

	  /**
	   * Sets the message
	   * 
	   * @param message the new message
	   */
	  public void setMessage(String message) {
	    this.message = message;
	  }

	  /**
	   * Gets the input
	   * 
	   * @return String
	   */
	  public String getInput() {
	    return input;
	  }

	  /**
	   * Sets the input
	   * 
	   * @param input the new input
	   */
	  public void setInput(String input) {
	    this.input = input;
	  }

	  /**
	   * Opens the dialog and returns the input
	   * 
	   * @return String
	   */
	  public String open(final boolean showEcho) {
	    // Create the dialog window
	    Shell shell = new Shell(getParent(), getStyle());
	    shell.setText(getText());
	    createContents(shell, showEcho);
	    shell.pack();
	    shell.open();
	    Display display = getParent().getDisplay();
	    while (!shell.isDisposed()) {
	      if (!display.readAndDispatch()) {
	        display.sleep();
	      }
	    }
	    // Return the entered value, or null
	    return input;
	  }

	  /**
	   * Creates the dialog's contents
	   * 
	   * @param shell the dialog window
	   */
	  private void createContents(final Shell shell, final boolean showEcho) {
	    shell.setLayout(new GridLayout(2, true));

	    // Show the message
	    Label label = new Label(shell, SWT.NONE);
	    label.setText(message);
	    GridData data = new GridData();
	    data.horizontalSpan = 2;
	    label.setLayoutData(data);

	    // Display the input box
	    final Text text = new Text(shell, SWT.BORDER);
	    if(input != null) {
	    	text.setText(input);
	    	text.setSelection(0, input.length());
	    }
	    if(!showEcho)
	    	text.setEchoChar('X');
	    data = new GridData(GridData.FILL_HORIZONTAL);
	    data.horizontalSpan = 2;
	    text.setLayoutData(data);

	    // Create the OK button and add a handler
	    // so that pressing it will set input
	    // to the entered value
	    Button ok = new Button(shell, SWT.PUSH);
	    ok.setText(l10n._("OK"));
	    data = new GridData(GridData.FILL_HORIZONTAL);
	    ok.setLayoutData(data);
	    ok.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	        input = text.getText();
	        shell.close();
	      }
	    });

	    // Create the cancel button and add a handler
	    // so that pressing it will set input to null
	    Button cancel = new Button(shell, SWT.PUSH);
	    cancel.setText(l10n._("Cancel"));
	    data = new GridData(GridData.FILL_HORIZONTAL);
	    cancel.setLayoutData(data);
	    cancel.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	        input = null;
	        shell.close();
	      }
	    });

	    // Set the OK button as the default, so
	    // user can type input and press Enter
	    // to dismiss
	    shell.setDefaultButton(ok);
	  }
	  
	  
    @Override
    public String userInputPassword(String subject, String message) {
        setText(subject);
        setMessage(message);
        this.input = "";
        return open(false);
    }

    @Override
    public String userInputText(String subject, String message) {
        setText(subject);
        setMessage(message);
        this.input = "";
        return open(true);
    }
}