/**
 * 
 */
package de.ibh.StrongRoom.Client.SWT;

import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;

import org.bouncycastle.openpgp.PGPPublicKey;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

import de.ibh.RandomArt.RAImage;
import de.ibh.StrongRoom.Client.Crypto;
import de.ibh.StrongRoom.Crypto.IPassphraseRequest;
import de.ibh.StrongRoom.l10n.StrongRoomL10N;


/**
 * @author ellguth
 *
 */
public class PassphraseDialog implements IPassphraseRequest {
	private static PassphraseDialog self = null;
	private String input;
	private PGPPublicKey lastPubKey = null;
	private static int timeout = 7200;
	private Timer timer = null;
	private boolean remember = true;
	private StrongRoomL10N l10n = StrongRoomL10N.getInstance();

	public static PassphraseDialog get() {
		if(self == null)
			self = new PassphraseDialog();
		
		return self;
	}
	
	public static void setInputTimeout(int timeout) {
		PassphraseDialog.timeout = timeout;
	}
	
	protected PassphraseDialog() {
		super();
	}

	public synchronized void clearInput() {
		input = null;
        System.gc();
	}
	
	@SuppressWarnings("unchecked")
	private void createContent(final Shell shell, final PGPPublicKey pubKey) {
		shell.setLayout(new GridLayout(2, false));

		final String fpr = Crypto.formatFPR(pubKey.getFingerprint());
		
	    /* label */
	    Label label = new Label(shell, SWT.NONE);
	    label.setText(l10n._("You need to enter the passphrase to unlock the following secret key to proceed."));
	    label.setLayoutData(new GridData(SWT.LEFT, SWT.TOP, true, false, 2, 1));

	    
	    /* FPR */
        Group fprGroup = new Group(shell, SWT.NONE);
        fprGroup.setText(l10n._("Key"));
        fprGroup.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));
        fprGroup.setLayout(new GridLayout(3, false));
	    
	    /*  Random Art  */
		Canvas canvasFPR = new Canvas(fprGroup, SWT.NONE);
		canvasFPR.addPaintListener(new PaintListener() {
			public void paintControl(PaintEvent e) {
				final Image img = RAImage.create(fpr, shell.getDisplay());
				e.gc.drawImage(img, 0, 0);
			}
		});
		GridData gd = new GridData(SWT.CENTER, SWT.CENTER, false, false, 1, 2);
		gd.heightHint = RAImage.getDefaultHeight();
		gd.minimumHeight = RAImage.getDefaultHeight();
		gd.widthHint = RAImage.getDefaultWidth();
		gd.minimumWidth = RAImage.getDefaultWidth();
		canvasFPR.setLayoutData(gd);
	    
	    /*  user ID  */
		label = new Label(fprGroup, SWT.NONE);
	    label.setText("UID:");
	    label.setLayoutData(new GridData(SWT.END, SWT.CENTER, false, false));

		label = new Label(fprGroup, SWT.NONE | SWT.BOLD);
	    Iterator<String> it = pubKey.getUserIDs();
	    if(it.hasNext())
		    label.setText(GUI.escapeStr(it.next()));
		else
		    label.setText("-");
	    label.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));

	    
	    /*  FPR  */
		label = new Label(fprGroup, SWT.NONE);
	    label.setText("FPR:");
	    label.setLayoutData(new GridData(SWT.END, SWT.CENTER, false, false));

		label = new Label(fprGroup, SWT.NONE | SWT.BOLD);
	    label.setText(fpr);
	    label.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));

	    
	    /* inputGroup */
	    
        Group inputGroup = new Group(shell, SWT.NONE);
        inputGroup.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));
        inputGroup.setLayout(new GridLayout(3, false));
        inputGroup.setText(l10n._("Passphrase"));
        
	    /* input box */
		final Canvas inputCanvas = new Canvas(inputGroup, SWT.NONE);
		gd = new GridData(SWT.CENTER, SWT.CENTER, false, false, 1, 2);
		gd.heightHint = RAImage.getDefaultHeight();
		gd.minimumHeight = RAImage.getDefaultHeight();
		gd.widthHint = RAImage.getDefaultWidth();
		gd.minimumWidth = RAImage.getDefaultWidth();
		inputCanvas.setLayoutData(gd);
        
		final Text text = new Text(inputGroup, SWT.BORDER);
	    text.setEchoChar('x');
	    text.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));

	    final Button savePwdButton = new Button(inputGroup, SWT.CHECK);
	    savePwdButton.setText(l10n._("remember passphrase for ") + timeout + l10n._(" seconds"));
	    savePwdButton.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));
	    savePwdButton.setSelection(remember);
    	    
	    
	    /* OK button */
	    final Button ok = new Button(shell, SWT.PUSH);
	    ok.setText(l10n._("OK"));
	    ok.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
	    ok.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	        input = text.getText();
	        remember = savePwdButton.getSelection();
	        shell.close();
	      }
	    });
	    ok.setImage(shell.getDisplay().getSystemImage(SWT.OK));
	    shell.setDefaultButton(ok);

	    
	    /* cancel button */
	    Button cancel = new Button(shell, SWT.PUSH);
	    cancel.setImage(shell.getDisplay().getSystemImage(SWT.CANCEL));
	    cancel.setText(l10n._("Cancel"));
	    cancel.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
	    cancel.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	        input = null;
	        shell.close();
	      }
	    });

	    
	    /* add listener */
		inputCanvas.addPaintListener(new PaintListener() {
			public void paintControl(PaintEvent e) {
            	String s = text.getText();
            	String seed = null;

            	if(!s.equals("")) {
            		seed = "";

            		for(int i=0; i<s.length(); i++) {
	            		seed += Integer.toHexString(s.charAt(i) % 4);
	            	}
            	}
            	s = null;
            	
				e.gc.drawImage(RAImage.create(seed, shell.getDisplay()), 0, 0);
			}
		});

		text.addListener(SWT.Modify, new Listener() {
            @Override
            public void handleEvent(Event arg0) {
                inputCanvas.redraw();
                ok.setEnabled(!text.getText().equals(""));
                shell.forceActive();
            }
        });

        if(input != null) {
	    	text.setText(input);
	    	text.setSelection(0, input.length());
	    }
        ok.setEnabled(!text.getText().equals(""));
	}
  
	public String getPassphrase(final String purpose, final boolean userMustConfirm, final PGPPublicKey pubKey, final Object userData) {
		Shell parent = (Shell) userData;

	    Shell shell = new Shell(parent, SWT.DIALOG_TRIM | SWT.APPLICATION_MODAL);

	    if(lastPubKey != null &&
	    		!Crypto.formatFPR(lastPubKey.getFingerprint()).equals(Crypto.formatFPR(pubKey.getFingerprint()))) {
	    	input = null;
	    	System.gc();
	    }
	    
	    if(!userMustConfirm && lastPubKey != null && input != null &&
	    		Crypto.formatFPR(lastPubKey.getFingerprint()).equals(Crypto.formatFPR(pubKey.getFingerprint()))) {
			
	    	String ret = input;

	    	if(ret != null)
				return ret;
	    }
		
		if(timer != null) {
			timer.cancel();
			timer.purge();
		}
	    
	    shell.setText(purpose);
	    createContent(shell, pubKey);
	    shell.pack();
	    shell.open();
	    Display display = parent.getDisplay();
	    while (!shell.isDisposed()) {
	      if (!display.readAndDispatch()) {
	        display.sleep();
	      }
	    }
	    
	    if(input == null)
	    	return null;
	  
	    lastPubKey = pubKey;

	    if(remember) {
	    	timer = new Timer("PassphraseDialog.clearInput");
	    	timer.schedule(new ClearInput(), timeout * 1000);

	    	return input;
	    }

	    String ret = input;
	    input = null;
	    
	    return ret;
	}
	
	class ClearInput extends TimerTask {
		@Override
		public void run() {
			clearInput();
		}
	}
}
