package de.ibh.StrongRoom.Client.SWT;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

import de.ibh.StrongRoom.Client.Crypto;
import de.ibh.StrongRoom.Client.Config.BaseConf;
import de.ibh.StrongRoom.Client.Lucene.SRSearch;
import de.ibh.StrongRoom.Client.Lucene.SearchResult;
import de.ibh.StrongRoom.DB.DBConn;
import de.ibh.StrongRoom.DB.Role;
import de.ibh.StrongRoom.DB.UserRole;
import de.ibh.StrongRoom.Exception.SRGenericException;
import de.ibh.StrongRoom.Exception.SRRoleNotFoundException;
import de.ibh.StrongRoom.l10n.StrongRoomL10N;

public class SearchDetail {
	Shell shell;
    private StrongRoomL10N l10n = StrongRoomL10N.getInstance();	

	public SearchDetail(final Shell s, final DBConn dbconn, final Crypto crypto, final BaseConf baseconf) throws SRGenericException {
		shell = new Shell(s);

		shell.setText(l10n._("Search by details"));
		shell.setLayout(new GridLayout(2, false));
		
		Label queryLabel = new Label(shell, SWT.NONE);
		queryLabel.setText(l10n._("Query:"));
		queryLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
		
		final Text queryText = new Text(shell, SWT.BORDER);
		queryText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));

		Label rolesLabel = new Label(shell, SWT.NONE);
		rolesLabel.setText(l10n._("Roles:"));
		rolesLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
		
		final Combo rolesCombo = new Combo(shell, SWT.DROP_DOWN | SWT.BORDER | SWT.READ_ONLY);
		rolesCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		rolesCombo.setText(l10n._("<all>"));
		rolesCombo.add(l10n._("<all>"));
		rolesCombo.select(0);

		final List<UserRole> uroles = dbconn.getUserRolesByFPR(baseconf.getSecretKeyFPR());
		final Map<Long, Role> roles = dbconn.getRoles();
		
		final Map<Integer, Role> comboMap = new HashMap<Integer, Role>();
		for(UserRole urole : uroles) {
			Role role = roles.get(urole.getRoleID());
			
			if(role == null)
				throw new SRRoleNotFoundException(urole.getRoleID());
			
			comboMap.put(rolesCombo.getItemCount(), role);
			rolesCombo.add(role.getName());
		}
		rolesCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
	    
	    /* OK button */
	    final Button searchButton = new Button(shell, SWT.PUSH);
	    searchButton.setText(l10n._("&Search"));
	    searchButton.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
	    searchButton.setImage(shell.getDisplay().getSystemImage(SWT.OK));
	    shell.setDefaultButton(searchButton);

	    /* Close button */
	    Button closeButton = new Button(shell, SWT.PUSH);
	    closeButton.setImage(shell.getDisplay().getSystemImage(SWT.CANCEL));
	    closeButton.setText(l10n._("&Close"));
	    closeButton.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
	    closeButton.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	        shell.close();
	      }
	    });

        searchButton.addSelectionListener(new SelectionAdapter() {
		      public void widgetSelected(SelectionEvent event) {
		    	  final int sel = rolesCombo.getSelectionIndex();

		    	  try {
		    		  final Map<Role,List<SearchResult>> tds = new HashMap<Role,List<SearchResult>>();
		    		  
			    	  if(sel == 0) {
		    			  for(Role role : comboMap.values()) {
					    	  tds.put(role, SRSearch.searchIndex(role, dbconn, crypto, shell, queryText.getText(), 10));
			    		  }
			    	  }
			    	  else {
			    		  tds.put(comboMap.get(sel), SRSearch.searchIndex(comboMap.get(sel), dbconn, crypto, shell, queryText.getText(), 10));
			    	  }
			    	  
			    	  new SearchDetailResults(s, dbconn, crypto, baseconf, tds);
			    	  
			    	  shell.close();
		    	  } catch (Exception e) {
		    		  // TODO Auto-generated catch block
		    		  e.printStackTrace();
		    	  }
		      }
		});

		shell.pack();
		shell.open();
	    Display display = shell.getDisplay();
	    while (!shell.isDisposed()) {
	    	if (!display.readAndDispatch())
	    		display.sleep();
	    }
	    shell.dispose();
	}
}
