package de.ibh.StrongRoom.Client.SWT;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;

import de.ibh.StrongRoom.Client.Crypto;
import de.ibh.StrongRoom.Client.Config.BaseConf;
import de.ibh.StrongRoom.DB.DBConn;
import de.ibh.StrongRoom.DB.Role;
import de.ibh.StrongRoom.DB.UserRole;
import de.ibh.StrongRoom.Exception.SRGenericException;
import de.ibh.StrongRoom.Exception.SRRoleNotFoundException;
import de.ibh.StrongRoom.l10n.StrongRoomL10N;

public class SearchIndexer {
    private StrongRoomL10N l10n = StrongRoomL10N.getInstance();
    
	public SearchIndexer(final Shell s, final DBConn dbconn, final BaseConf baseconf, final Crypto crypto) throws SRGenericException {
		final Shell shell = new Shell(s, SWT.DIALOG_TRIM | SWT.APPLICATION_MODAL);

		shell.setText(l10n._("Build search index"));
		shell.setLayout(new GridLayout(2, false));
		
		final Label rolesLabel = new Label(shell, SWT.NONE);
		rolesLabel.setText(l10n._("Roles:"));
		rolesLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
		
		final Combo rolesCombo = new Combo(shell, SWT.DROP_DOWN | SWT.BORDER | SWT.READ_ONLY);
		rolesCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		rolesCombo.setText(l10n._("<any>"));
		rolesCombo.add(l10n._("<any>"));
		rolesCombo.select(0);
        
		final List<UserRole> uroles = dbconn.getUserRolesByFPR(baseconf.getSecretKeyFPR());
		final Map<Long, Role> roles = dbconn.getRoles();
		
		final Map<Integer, Role> comboMap = new HashMap<Integer, Role>();
		for(UserRole urole : uroles) {
			Role role = roles.get(urole.getRoleID());
			
			if(role == null)
				throw new SRRoleNotFoundException(urole.getRoleID());
			
			comboMap.put(rolesCombo.getItemCount(), role);
			rolesCombo.add(role.getName());
		}
		rolesCombo.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));

	    
	    /* OK button */
	    final Button buildButton = new Button(shell, SWT.PUSH);
	    buildButton.setText(l10n._("&Build index"));
	    buildButton.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
	    buildButton.setImage(shell.getDisplay().getSystemImage(SWT.OK));
	    shell.setDefaultButton(buildButton);

	    
	    /* Close button */
	    Button closeButton = new Button(shell, SWT.PUSH);
	    closeButton.setImage(shell.getDisplay().getSystemImage(SWT.CANCEL));
	    closeButton.setText(l10n._("&Close"));
	    closeButton.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
	    closeButton.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	        shell.close();
	      }
	    });
		
	    buildButton.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	    	  final int sel = rolesCombo.getSelectionIndex();

	    	  try {
		    	  if(sel == 0) {
		    		  for(Role role : comboMap.values()) {
		    			  new SearchIdxBuilder(shell, dbconn, baseconf, crypto, role);
		    		  }
		    	  }
		    	  else {
		    		  new SearchIdxBuilder(shell, dbconn, baseconf, crypto, comboMap.get(sel));
		    	  }
		    	  
		    	  MessageBox mb = new MessageBox(shell);
		    	  mb.setText(l10n._("Indexing completed!"));
		    	  mb.setMessage(l10n._("The indexing has been successfully completed!"));
		    	  mb.open();
		    	  
		    	  shell.close();
	    	  } catch (Exception e) {
	    		  // TODO Auto-generated catch block
	    		  e.printStackTrace();
	    	  }
	      }
	    });
        
        
		shell.pack();
		shell.open();
	    Display display = shell.getDisplay();
	    while (!shell.isDisposed()) {
	    	if (!display.readAndDispatch())
	    		display.sleep();
	    }
	    shell.dispose();
	}
}
