/**
 * 
 */
package de.ibh.StrongRoom.Client.SWT;


import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.eclipse.swt.*;
import org.eclipse.swt.custom.*;
import org.eclipse.swt.events.DragDetectEvent;
import org.eclipse.swt.events.DragDetectListener;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseTrackListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

import bmsi.util.Diff;
import bmsi.util.DiffPrint.Base;
import bmsi.util.DiffPrint.ContextPrint;
import bmsi.util.DiffPrint.EdPrint;
import bmsi.util.DiffPrint.NormalPrint;
import bmsi.util.DiffPrint.UnifiedPrint;

import de.ibh.StrongRoom.Client.Crypto;
import de.ibh.StrongRoom.Client.Misc;
import de.ibh.StrongRoom.Crypto.IRetrievePublicKeyRingHelper;
import de.ibh.StrongRoom.Crypto.IRetrieveSecretKeyRingHelper;
import de.ibh.StrongRoom.Crypto.CryptoBackend.ECryptoPGP_KeyType;
import de.ibh.StrongRoom.DB.CryptoData;
import de.ibh.StrongRoom.DB.CryptoEntry;
import de.ibh.StrongRoom.DB.Customer;
import de.ibh.StrongRoom.DB.DBConn;
import de.ibh.StrongRoom.DB.Role;
import de.ibh.StrongRoom.DB.UserRole;
import de.ibh.StrongRoom.Exception.SRDBSQLException;
import de.ibh.StrongRoom.Exception.SRGenericCryptoEngineException;
import de.ibh.StrongRoom.Exception.SRGenericException;
import de.ibh.StrongRoom.Exception.SRPublicKeyNotFoundException;
import de.ibh.StrongRoom.Exception.SRRoleNotFoundException;
import de.ibh.StrongRoom.Exception.SRSecretKeyNotFoundException;
import de.ibh.StrongRoom.Exception.SRSigningKeyNotConfirmedException;
import de.ibh.StrongRoom.Log.DebugMessage;
import de.ibh.StrongRoom.Log.ErrorMessage.EMessageTypes;
import de.ibh.StrongRoom.l10n.StrongRoomL10N;

/**
 * @author ellguth
 *
 */
public class TabFolders {
    private enum ETITabType {
        SEARCH,
        CRYPTO,
        NEW,
    }

    private class CryptoTab {
        ETITabType type;
        private Customer cu;
        private CryptoEntry ce;
        private int tbMask = -1;
        private int menuMask = -1;
        private HashMap<Integer, Listener []> tbListeners;
        private HashMap<Integer, Listener []> menuListeners;
        
        protected CryptoTab() {
            this.type = ETITabType.SEARCH;
        }
        
        protected CryptoTab(final Customer cu) {
            this.type = ETITabType.NEW;
            this.cu = cu;
        }
        
        protected CryptoTab(final CryptoEntry ce) {
            this.type = ETITabType.CRYPTO;
            this.cu = ce.getCustomer();
            this.ce = ce;
        }
        
        protected ETITabType getType() {
            return type;
        }
        
        protected Customer getCustomer() {
            if(type == ETITabType.SEARCH)
                throw new IllegalStateException();

            return cu;
        }
        
        protected CryptoEntry getCryptoEntry() {
            if(type != ETITabType.CRYPTO)
                throw new IllegalStateException();

            return ce;
        }
        
        protected void saveMenuState() {
            if(toolbar == null) return;
            tbListeners = toolbar.getItemListeners();
            tbMask = toolbar.getItemMask();
            
            menuListeners = mainmenu.getItemListeners();
            menuMask = mainmenu.getItemMask();
        }
        
        protected boolean loadMenuState() {
            if(toolbar == null || mainmenu == null || tbMask == -1 || menuMask == -1) return false;
            toolbar.removeAllItemListeners();
            toolbar.setItemListeners(tbListeners);
            toolbar.disableAllItems();
            toolbar.enableItems(tbMask);
            
            mainmenu.removeAllItemListeners();
            mainmenu.setItemListeners(menuListeners);
            mainmenu.disableAllItems();
            mainmenu.enableItems(menuMask);
           
            return true;
        }
    }
    
    private CTabFolder      tabFolder;
    private Map<CryptoTab, CTabItem> tabs      = new HashMap<CryptoTab, CTabItem>();
    private Font           monospace;
    private Font           headline;
    private Toolbar        toolbar;
    private MainMenu       mainmenu;
    private Composite      parentComposite;
    
    private ErrorMsgDialog errmsg;
    private Map<String, String> roleKeys = new HashMap<String, String>();
    
    private Vector<StyleRange> cachedStyles = new Vector<StyleRange>();

    private final GUI gui;
    
    private final String   tcTextKey           = "text";
    private final String   tcVersionListKey    = "versionlist";
    private final String   tcUnsavedEntry      = "unsaved";
    private final String   tcLabelChangedUID   = "changed_uid";
    private final String   tcLabelChangedFPR   = "changed_fpr";
    private final String   tcLabelChangedTimestamp = "changed_ts";
    private final String   tcCryptoData        = "cd_";
    private final String   tcDiffView          = "diff";
    private final String   tcToolbarDiff       = "tb_diff";
    private final String   tcSearchBarVisible  = "tb_searchbar";
    private final Color    bgColorTextCrypto   = new Color(Display.getDefault(), 255, 224, 230);
    private final StrongRoomL10N l10n = StrongRoomL10N.getInstance();  
    
    static DebugMessage logger = DebugMessage.getInstance();
    
    @SuppressWarnings("deprecation")
    public TabFolders(final Composite parent, final GUI gui, final int style) {
        this.gui = gui;
        
        tabFolder = new CTabFolder(parent, style);

        CTabItem ti = new CTabItem(tabFolder, SWT.NONE);
        ti.setText(l10n._("Quick Search"));
        ti.setImage(GUI.getIcon(parent.getDisplay(), "search.png"));
        tabs.put(new CryptoTab(), ti);
        tabFolder.setSelection(ti);
        tabFolder.setSimple(false);
        ti.setShowClose(false);
        
        tabFolder.pack();
        parentComposite = parent;
        errmsg = ErrorMsgDialog.getInstance(parentComposite.getShell());        
        
        toolbar = Toolbar.getInstance(parent);
        mainmenu = MainMenu.getInstance(parent.getShell());
        
        tabFolder.addListener(SWT.MenuDetect, new Listener() {
            public void handleEvent(Event event) {
                int iboundx = event.x - tabFolder.getParent().toDisplay(tabFolder.getBounds().x, tabFolder.getBounds().y).x;
                int iboundy = event.y - tabFolder.getParent().toDisplay(tabFolder.getBounds().x, tabFolder.getBounds().y).y;
                
                CTabItem ti = tabFolder.getItem(new Point(iboundx, iboundy));
                
                if(ti != null) {
                    showCryptoTabContextMenu((Shell) tabFolder.getParent(), tabFolder.getDisplay(), ti, event);
                }
            }
        });
        
        tabFolder.addListener(SWT.Paint, new Listener() {
                private CryptoTab lastTab = null;
            
                public void handleEvent(Event event) {
                    final CTabItem  tis  = tabFolder.getSelection();
                    final CryptoTab ct = getCryptoTab(tis);

                    if(ct == null || lastTab == ct)
                        return;
                    
                    if(lastTab != null) {
                        // Save the state of the last crypto tab.
                        lastTab.saveMenuState();
                    } else
                        getCryptoTab(getSearchTab()).saveMenuState();
                    
                    if(ct.getType() == ETITabType.SEARCH) {
                        setSearchTextEnabled(true);
                        
                        ct.loadMenuState();
                        toolbar.disableItems(Toolbar.ITEM_CLOSE | Toolbar.ITEM_EDIT | Toolbar.ITEM_SAVE);
                        mainmenu.disableItems(MainMenu.ITEM_CLOSE | MainMenu.ITEM_CLOSETAB | MainMenu.ITEM_SAVE | MainMenu.ITEM_COPY | MainMenu.ITEM_CUT | MainMenu.ITEM_PASTE);
                        getSearchTab().getControl().setFocus();
                    } else if(ct.getType() == ETITabType.NEW) {
                        setSearchTextEnabled(false);
                        
                        if(!ct.loadMenuState()) {
                            toolbar.enableItems(Toolbar.ITEM_CLOSE | Toolbar.ITEM_SAVE);
                            toolbar.disableItems(Toolbar.ITEM_EDIT | Toolbar.ITEM_NEW | Toolbar.ITEM_OPEN);
                            mainmenu.enableItems(MainMenu.ITEM_CLOSE | MainMenu.ITEM_CLOSETAB | MainMenu.ITEM_SAVE);
                            mainmenu.disableItems(MainMenu.ITEM_OPEN | MainMenu.ITEM_COPY | MainMenu.ITEM_CUT | MainMenu.ITEM_PASTE);
                            
                            Listener ctiListener = new Listener() {
                                public void handleEvent(Event e) {
                                    removeCTabItem(tabs.get(ct));
                                }
                            };
                            
                            toolbar.removeItemListener(Toolbar.ITEM_CLOSE);
                            toolbar.addItemListener(Toolbar.ITEM_CLOSE, ctiListener);
                            mainmenu.removeItemListener(MainMenu.ITEM_CLOSE);
                            mainmenu.addItemListener(MainMenu.ITEM_CLOSE, ctiListener);
                            mainmenu.removeItemListener(MainMenu.ITEM_CLOSETAB);
                            mainmenu.addItemListener(MainMenu.ITEM_CLOSETAB, ctiListener);                            
                        }
                    } else if(ct.getType() == ETITabType.CRYPTO) {
                        setSearchTextEnabled(false);
                        
                        if(!ct.loadMenuState()) {
                            Listener ctiListener = new Listener() {
                                public void handleEvent(Event e) {
                                	final CTabItem ti = tabs.get(ct);
                                	
                                	if((e.widget == mainmenu.getItem(MainMenu.ITEM_CLOSE) || 
                                	        e.widget == mainmenu.getItem(MainMenu.ITEM_CLOSETAB)) &&
                                			ti.getData(tcSearchBarVisible) != null) {

                                		mainmenu.getItemListeners().get(MainMenu.ITEM_FIND)[0].handleEvent(e);
                                		
                                		return;
                                	}

                                	removeCTabItem(ti);
                                }
                            };
                            
                            Listener editListener = new Listener() {
                                public void handleEvent(Event e) {
                                    editCryptoEntry(tabs.get(ct));
                                }
                            };
                            
                            toolbar.disableItems(Toolbar.ITEM_OPEN | Toolbar.ITEM_NEW);
                            
                            if(getTextCrypto(tabs.get(ct)).isEnabled() &&
                            	!getTextCrypto(tabs.get(ct)).getEditable() &&
                            	!ct.ce.isReadonly()) {
                                toolbar.enableItems(Toolbar.ITEM_EDIT);
                                toolbar.addItemListener(Toolbar.ITEM_EDIT, editListener);
                            }
                            toolbar.enableItems(Toolbar.ITEM_CLOSE);
                            // Reset the item listener for the toolbar.
                            toolbar.removeItemListener(Toolbar.ITEM_CLOSE);
                            toolbar.addItemListener(Toolbar.ITEM_CLOSE, ctiListener);
                            
                            // Reset the item listener for the mainmenu.
                            mainmenu.disableItems(MainMenu.ITEM_OPEN);
                            mainmenu.enableItems(MainMenu.ITEM_CLOSE | MainMenu.ITEM_CLOSETAB | MainMenu.ITEM_COPY | MainMenu.ITEM_CUT | MainMenu.ITEM_PASTE);
                            mainmenu.removeItemListener(MainMenu.ITEM_CLOSE);
                            mainmenu.addItemListener(MainMenu.ITEM_CLOSE, ctiListener);
                            mainmenu.removeItemListener(MainMenu.ITEM_CLOSETAB);
                            mainmenu.addItemListener(MainMenu.ITEM_CLOSETAB, ctiListener);                            
                        }
                    }

                    lastTab = ct;
                }
            }
        );
        
        tabFolder.addCTabFolderListener(new CTabFolderAdapter() {
            public void itemClosed(CTabFolderEvent event) {
                if(event.item instanceof CTabItem)
                    removeCTabItem((CTabItem) event.item);
                event.doit = false;
            }
        });

        tabFolder.addDragDetectListener(new DragDetectListener() {
            
            @Override
            public void dragDetected(DragDetectEvent event) {
                parentComposite.setCursor(new Cursor(parentComposite.getShell().getDisplay(),
                        SWT.CURSOR_HAND));
            }
        });
        tabFolder.addMouseListener(new MouseListener() {
            
            @Override
            public void mouseUp(MouseEvent arg0) {
                parentComposite.setCursor(new Cursor(parentComposite.getShell().getDisplay(),
                        SWT.CURSOR_ARROW));                
            }
            
            @Override
            public void mouseDown(MouseEvent arg0) {
                // TODO Auto-generated method stub
                
            }
            
            @Override
            public void mouseDoubleClick(MouseEvent arg0) {
                // TODO Auto-generated method stub
                
            }
        });
        
        tabFolder.addMouseTrackListener(new MouseTrackListener() {
            
            @Override
            public void mouseHover(MouseEvent event) {
                if((event.stateMask & SWT.BUTTON1) != SWT.BUTTON1)
                    return;

                final CTabItem sti = tabFolder.getSelection();
                final CTabItem dti = tabFolder.getItem(new Point(event.x, event.y));
                if(dti == null || sti == dti)
                    return;

                swapTabs(sti, dti);
            }
            
            @Override
            public void mouseExit(MouseEvent event) {
                // TODO Auto-generated method stub
                
            }
            
            @Override
            public void mouseEnter(MouseEvent event) {
                // TODO Auto-generated method stub
                
            }
        });
        
        mainmenu.addItemListener(MainMenu.ITEM_COPY, new Listener() {
            
            @Override
            public void handleEvent(Event event) {
                final CTabItem  ti = tabFolder.getSelection();
                final CryptoTab ct = getCryptoTab(ti);
                
                if(ct.getType() == ETITabType.CRYPTO) {
                    final StyledText text = getTextCrypto(ti);
                    if(text == null)
                        return;

                    text.copy();
                }
            }
        });
        
        mainmenu.addItemListener(MainMenu.ITEM_CUT, new Listener() {
            
            @Override
            public void handleEvent(Event event) {
                final CTabItem  ti = tabFolder.getSelection();
                final CryptoTab ct = getCryptoTab(ti);
                
                if(ct.getType() == ETITabType.CRYPTO) {
                    final StyledText text = getTextCrypto(ti);
                    if(text == null || text.getEditable() == false)
                        return;
                    
                    text.cut();
                }
            }
        });        
        
        mainmenu.addItemListener(MainMenu.ITEM_PASTE, new Listener() {
            
            @Override
            public void handleEvent(Event event) {
                final CTabItem  ti = tabFolder.getSelection();
                final CryptoTab ct = getCryptoTab(ti);
                
                if(ct.getType() == ETITabType.CRYPTO) {
                    final StyledText text = getTextCrypto(ti);
                    if(text == null || text.getEditable() == false)
                        return;
                    
                    text.paste();
                }
            }
        });  
        
        
        monospace = new Font(tabFolder.getDisplay(), "Courier", 12, SWT.NONE);
        headline = new Font(tabFolder.getDisplay(), tabFolder.getFont().toString(), 16, SWT.BOLD);
    }
    
  
    public void setDefault(Control ctrlDefault) {
        if(ctrlDefault != null) getSearchTab().setControl(ctrlDefault);
    }
    
    public void setLayoutData(Object layoutData) {
        tabFolder.setLayoutData(layoutData);
    }
    
    
    public boolean tryToCloseAllTabs() {
        HashMap<CryptoTab, CTabItem> tabsCopy = new HashMap<CryptoTab, CTabItem>(tabs);
        for(Iterator<CryptoTab> it = tabsCopy.keySet().iterator(); it.hasNext();) {
            CryptoTab ct = it.next();
            if(ct.getType() != ETITabType.SEARCH) {
                tabFolder.setSelection(tabs.get(ct));
                removeCTabItem(tabs.get(ct));
            }
        }
        
        // Number of minimal tabs is 1 (search tab).
        return tabs.size() == 1 ? true : false;
    }
    
    
    public CTabFolder getParent() {
        return tabFolder;
    }
    
    
    private void swapTabs(CTabItem sti, CTabItem dti) {
        if(getCryptoTab(dti) == null || getCryptoTab(sti) == null || getCryptoTab(dti).getType() == ETITabType.SEARCH || getCryptoTab(sti).getType() == ETITabType.SEARCH)
            return;

        CryptoTab sCt = getCryptoTab(sti);
        CryptoTab dCt = getCryptoTab(dti);
        
        // Swap the hashmap
        tabs.remove(sCt);
        tabs.remove(dCt);
        
        Control sCtrl = sti.getControl();
        Control dCtrl = dti.getControl();
        String  sText = sti.getText();
        String  dText = dti.getText();
        Image   sImage = sti.getImage();
        Image   dImage = dti.getImage();
        
        // Swap the data.
        sti.setText(dText);
        dti.setText(sText);
        
        sti.setControl(dCtrl);
        dti.setControl(sCtrl);
        
        sti.setImage(dImage);
        dti.setImage(sImage);

        Object sData = sti.getData();
        Object dData = dti.getData();
        sti.setData(dData);
        dti.setData(sData);
        
        final String [] keyArr = new String [] {
                tcCryptoData, 
                tcDiffView,
                tcLabelChangedFPR,
                tcLabelChangedTimestamp,
                tcLabelChangedUID,
                tcSearchBarVisible, 
                tcTextKey, 
                tcToolbarDiff,
                tcUnsavedEntry,
                tcVersionListKey };
        
        for(String key : keyArr) {
            sData = sti.getData(key);
            dData = dti.getData(key);
            sti.setData(key, dData);
            dti.setData(key, sData);
        }
        
        tabs.put(dCt, sti);
        tabs.put(sCt, dti);

        tabFolder.setSelection(dti);        
    }
    
    
    private void setSearchTextEnabled(boolean active) {
        gui.setSearchTextEnabled(active);
    }
    
    
    private CryptoTab getCryptoTab(CTabItem ti) {
        for(Iterator<CryptoTab> it = tabs.keySet().iterator(); it.hasNext();) {
            CryptoTab ct = it.next();
            if(tabs.get(ct) == ti)
                return ct;
        }
        return null;
    }
    
    
    private CryptoTab getCryptoTab(CryptoEntry ce) {
        for(Iterator<CryptoTab> it = tabs.keySet().iterator(); it.hasNext();) {
            CryptoTab ct = it.next();
            if(ct.getType() == ETITabType.CRYPTO && ct.getCryptoEntry() == ce)
                return ct;
        }    
        return null;
    }
    
    
    private CryptoTab getCryptoTab(Customer cu) {
        for(Iterator<CryptoTab> it = tabs.keySet().iterator(); it.hasNext();) {
            CryptoTab ct = it.next();
            if(ct.getType() == ETITabType.NEW && ct.getCustomer().getId() == cu.getId())
                return ct;
        }        
        return null;
    }    
    
    
    private CTabItem getSearchTab() {
        for(Iterator<CryptoTab> it = tabs.keySet().iterator(); it.hasNext();) {
            CryptoTab ct = it.next();
            if(ct.getType() == ETITabType.SEARCH)
                return tabs.get(ct);
        }
        
        throw new IllegalStateException(l10n._("Did not find the search tab!"));
    }
    
    
    private List getListCrypto(CTabItem ti) {
        if(ti == null || ti.getData(tcVersionListKey) == null) return null;
        
        List listCrypto = null;
        if(ti.getData(tcVersionListKey)  instanceof List)
            listCrypto = (List) ti.getData(tcVersionListKey);

        return listCrypto;
    }
    
    
    private StyledText getTextCrypto(CTabItem ti) {
        if((ti == null || ti.getData(tcTextKey) == null || (ti.getData(tcTextKey) instanceof StyledText) == false)) return null;
        StyledText textCrypto = (StyledText) ti.getData(tcTextKey);
        
        return textCrypto;
    }
    
    
    private void removeCTabItem(CTabItem ti) {
        if(ti != null && ti != getSearchTab()) {
            if(ti.getData(tcUnsavedEntry) != null) {
                CloseTabDialog ctd = new CloseTabDialog(parentComposite.getShell(), ti.getText());
                int r = ctd.open();
                if(r == CloseTabDialog.CTD_CLOSENOSAVE) {
                } else if (r == CloseTabDialog.CTD_SAVE) {
                    try {
                        saveCryptoData(ti);
                    } catch (Exception ex) {
                        logger.debug(ex);
                        return;
                    }
                } else {
                    return;
                }
            }
                
        	CryptoEntry ce = null;
        	
            HashMap<CryptoTab, CTabItem> tabsCopy = new HashMap<CryptoTab, CTabItem>(tabs);
            Iterator<Map.Entry<CryptoTab, CTabItem>> entries = tabsCopy.entrySet().iterator();
            
            while (entries.hasNext()) {
                Map.Entry<CryptoTab, CTabItem> entry = entries.next();
                
                if(entry.getValue() == ti) {
                	ce = entry.getKey().getType() == ETITabType.CRYPTO ? entry.getKey().getCryptoEntry() : null;

                	tabs.remove (entry.getKey());
                    ti.dispose();
                }
                
            }
            
            /* drop RoleKey if not used any more */
            if(ce != null && ce.getRKey_FPR() != null) {
            	for(Iterator<CryptoTab> it = tabs.keySet().iterator(); it.hasNext(); ) {
            	    CryptoTab ct = it.next();

            	    if(ct.getType() != ETITabType.CRYPTO)
            	        continue;
            	        
            	    String itfpr = ct.getCryptoEntry().getRKey_FPR();
            		if(itfpr.equals(ce.getRKey_FPR()))
            			return;
            	}
            	
            	roleKeys.remove(ce.getRKey_FPR());
            }
            
            /* rumble away the confidential objects by the GC */
            System.gc();
        }        
    }
    
    
    private void editCryptoEntry(CTabItem ti) {
        if(ti != null && ti != getSearchTab()) {
            // TODO: Do some checks.
            List listCrypto = getListCrypto(ti);
            if(listCrypto.getSelectionIndex() == 0) {
                enableTextItem(ti);
            }
        }
    }
    
    
    private void saveCryptoData(CTabItem ti) throws SRDBSQLException, SRSigningKeyNotConfirmedException, SRSecretKeyNotFoundException, SRPublicKeyNotFoundException, SRGenericCryptoEngineException {
        final CryptoTab ct = getCryptoTab(ti);
        
        if(ct.type != ETITabType.CRYPTO)
            return;

        StyledText styledText = (StyledText) ti.getData(tcTextKey);

        final Crypto crypto = gui.getCrypto();
        final CryptoData cd = new CryptoData(
                ct.getCryptoEntry(),
                null,
                null,
                crypto.getFPR(ECryptoPGP_KeyType.USER_KEY),
                null,
                null);
        
        TextContent tc = new TextContent(parentComposite.getDisplay());
        tc.setContent(styledText.getText());
        final String ft = tc.toFormatedText(styledText.getStyleRanges());
        
        crypto.encryptRow(cd, ft, new IRetrievePublicKeyRingHelper() {
			
			@Override
			public String getPublicKeyring(ECryptoPGP_KeyType kt, Object userData) {
				if(kt != ECryptoPGP_KeyType.ROLE_KEY)
					return null;

				return ct.ce.getRole().getRoleKey().getPublicKey();
			}
		}, parentComposite.getShell());
        crypto.signRow(cd, new IRetrieveSecretKeyRingHelper() {
			
			@Override
			public String getSecretKeyring(ECryptoPGP_KeyType kt, Object userData) {
				if(kt != ECryptoPGP_KeyType.ROLE_KEY)
					return null;

				try {
					return Misc.b2s(crypto.decrypt(ECryptoPGP_KeyType.USER_KEY, ct.getCryptoEntry().getRole().getRoleKey(), userData));
				} catch (SRGenericException e) {
		            errmsg.printErrorMessage("Decryption error: " + e.getLocalizedMessage(), EMessageTypes.TYPE_ERROR);

		            return null;
				}
			}
		}, parentComposite.getShell());

        final CryptoData ocd = (CryptoData) getListCrypto(ti).getData(tcCryptoData + "0");
        String ots = "1970-01-01 00:00:00";
        if(ocd != null)
        	ots = ocd.getDate().toString();

        DBConn.getInstance().saveCryptoData(ots, ct.getCryptoEntry(), cd);
        
        disableTextItem(ti);
        ti.setImage(GUI.getIcon(ti.getDisplay(), "folder_green_open.png"));

        refreshListCryptoItem(ct);
    }
    
    
    private void createCryptoEntry(final CryptoTab ct, final String subject, final String description, final Role role) throws SRSigningKeyNotConfirmedException, SRSecretKeyNotFoundException, SRDBSQLException {
        if(ct.type != ETITabType.NEW)
            return;

        if(subject.length() < 5) {
            errmsg.printErrorMessage(l10n._("The minimum length of the subject must be 5 characters."), EMessageTypes.TYPE_WARN);
            return;
        }
        final Crypto crypto = gui.getCrypto();
        final CryptoEntry ce = new CryptoEntry(
                -1,
                ct.getCustomer(),
                subject,
                description,
                role,
                role.getRoleKey().getKeyFPR(),
                false,
                false,
                false,
                crypto.getFPR(ECryptoPGP_KeyType.USER_KEY),
                null,
                null);
        
        try {
			crypto.signRow(ce, new IRetrieveSecretKeyRingHelper() {
			    
			    @Override
			    public String getSecretKeyring(ECryptoPGP_KeyType kt, Object userData) {
			        if(kt != ECryptoPGP_KeyType.ROLE_KEY)
			            return null;

			        try {
						return Misc.b2s(crypto.decrypt(ECryptoPGP_KeyType.USER_KEY, role.getRoleKey(), userData));
					} catch (SRGenericException e) {
			            errmsg.printErrorMessage(l10n._("Decryption error: ") + e.getLocalizedMessage(), EMessageTypes.TYPE_ERROR);

			            return null;
					}
			    }
			}, parentComposite.getShell());
		} catch (SRGenericCryptoEngineException e) {
            errmsg.printErrorMessage(l10n._("Failed to create signature: ") + e.getMessage(), EMessageTypes.TYPE_WARN);
            return;
		}

        DBConn.getInstance().createCryptoEntry(ce);
        tabs.get(ct).setData(tcUnsavedEntry, null);
    }
    
    
    private void showCryptoTabContextMenu(Shell shell, Display display, final CTabItem ti, Event event)
    {
        Menu popup = new Menu(shell, SWT.POP_UP);
        MenuItem item = new MenuItem(popup, SWT.PUSH);
        item.setText(l10n._("Close"));
        item.addListener(SWT.Selection, new Listener() {
          public void handleEvent(Event e) {
              removeCTabItem(ti);
          }
        });
        popup.setLocation(event.x, event.y);
        popup.setVisible(true);
        while (!popup.isDisposed() && popup.isVisible()) {
          if (!display.readAndDispatch())
            display.sleep();
        }
        popup.dispose();        
    }
    
    
    private void showCryptoTextContextMenu(Shell shell, Display display, final CryptoTab ct, final StyledText text, Event event)
    {
        Menu popup = new Menu(shell, SWT.POP_UP);
        
        MenuItem item = new MenuItem(popup, SWT.PUSH);
        item.setText(l10n._("Cut"));
        item.setEnabled(text.getEditable());
        item.addListener(SWT.Selection, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
                text.cut();
            }
        });

        item = new MenuItem(popup, SWT.PUSH);
        item.setText(l10n._("Copy"));
        item.addListener(SWT.Selection, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
                text.copy();
            }
        });

        item = new MenuItem(popup, SWT.PUSH);
        item.setText(l10n._("Paste"));
        item.setEnabled(text.getEditable());
        item.addListener(SWT.Selection, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
                text.paste();
            }
        });

        new MenuItem(popup, SWT.SEPARATOR);
        
        item = new MenuItem(popup, SWT.PUSH);
        item.setText(l10n._("Select all"));
        item.addListener(SWT.Selection, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
                text.selectAll();
            }
        });
        
        new MenuItem(popup, SWT.SEPARATOR);
        
        item = new MenuItem(popup, SWT.PUSH);
        item.setText(l10n._("Edit"));
        item.setEnabled(!text.getEditable() && !ct.ce.isReadonly());
        item.addListener(SWT.Selection, new Listener() {
            
            @Override
            public void handleEvent(Event arg0) {
                editCryptoEntry(tabs.get(ct));
            }
        });        
        

        popup.setLocation(event.x, event.y);
        popup.setVisible(true);
    }
    
    
    private boolean refreshListCryptoItem(CryptoTab ct)
    {
        CryptoEntry ce = ct.getCryptoEntry();
        CTabItem ti = tabs.get(ct);

        List listCrypto = getListCrypto(ti);        
        if(listCrypto == null)
        	return false;
        
        DBConn db = DBConn.getInstance();
        if(db == null)
        	return false;

        /* free old data */
        for(int i = 0; i < listCrypto.getItemCount(); i++) {
        	listCrypto.setData(Integer.toString(i), null);
        }
        listCrypto.removeAll();
        
        int idx = 0;
        for(CryptoData cd : db.getCryptoData(ce)) {
        	listCrypto.add(gui.getUIDbyFPR(cd.getUKeyFPR()));
            listCrypto.setData(tcCryptoData + Integer.toString(idx), cd);
            idx++;
        }
        if(listCrypto.getItemCount() == 0)
            listCrypto.add("<empty>");
        
        listCrypto.select(0);
        return loadCryptoText(gui.getCrypto(), ct, tabFolder.getParent());        
    }
    
    
    private void enableSaveButton(final CTabItem ti) {
        // Check if the save button already enabled.
        if(ti.getData(tcUnsavedEntry) != null)
            return; 
        
        ti.setData(tcUnsavedEntry, true);
        ti.setImage(GUI.getIcon(ti.getDisplay(), "folder_red_open.png"));
        toolbar.enableItems(Toolbar.ITEM_SAVE);
        mainmenu.enableItems(MainMenu.ITEM_SAVE);
        
        Listener sceListener = new Listener() {
            public void handleEvent(Event e) {
                try {
                    saveCryptoData(ti);
                } catch (Exception ex) {
                    logger.debug(ex);
                }
            }
        };                

        mainmenu.removeItemListener(MainMenu.ITEM_SAVE);
        mainmenu.addItemListener(MainMenu.ITEM_SAVE, sceListener);
        
        toolbar.removeItemListener(Toolbar.ITEM_SAVE);
        toolbar.addItemListener(Toolbar.ITEM_SAVE, sceListener);        
    }
    
    
    private void enableTextItem(final CTabItem ti) {
        if(ti == null) return;

        final StyledText textCrypto = getTextCrypto(ti);
        if(textCrypto == null) return;

        if(textCrypto.getData() instanceof ModifyListener) {
            textCrypto.removeModifyListener((ModifyListener) textCrypto.getData());
            textCrypto.setData(null);            
        }
        textCrypto.setEditable(true);

        final ModifyListener li = new ModifyListener() {
            public void modifyText(ModifyEvent e) {
                enableSaveButton(ti);
            }
        };
        textCrypto.addModifyListener(li);
        textCrypto.setData(li);
        
        toolbar.disableItems(Toolbar.ITEM_EDIT);
        final ToolBar tbDiff = (ToolBar) ti.getData(tcToolbarDiff);
        tbDiff.setEnabled(false);
        
        toolbar.enableItems(GUI.T_STYLEELEMENTS);
        Listener styleListener = new Listener() {
            
            @Override
            public void handleEvent(Event event) {
                if(!(event.widget instanceof ToolItem || event.widget instanceof MenuItem) || !(event.widget.getData() instanceof Integer)) return;
                // widget.getData returns the Toolbar item ID.
                if(setStyle(textCrypto, (Integer) event.widget.getData()) == true)
                    enableSaveButton(ti);
                toolbar.setBoldItemChecked(false);
                toolbar.setItalicItemChecked(false);
                toolbar.setUnderlineItemChecked(false);
                toolbar.setColorItemChecked(false);
            }
        };
        toolbar.addItemListener(Toolbar.ITEM_BOLD, styleListener);
        toolbar.addItemListener(Toolbar.ITEM_ITALIC, styleListener);
        toolbar.addItemListener(Toolbar.ITEM_UNDERLINE, styleListener);
        toolbar.addItemListener(Toolbar.ITEM_BLACK, styleListener);
        toolbar.addItemListener(Toolbar.ITEM_RED, styleListener);
        toolbar.addItemListener(Toolbar.ITEM_GREEN, styleListener);
        toolbar.addItemListener(Toolbar.ITEM_BLUE, styleListener);
    }
    
    
    private void disableTextItem(CTabItem ti) {
        if(ti == null) return;
        
        StyledText textCrypto = getTextCrypto(ti);
        if(textCrypto == null) return;

        if(textCrypto.getData() instanceof ModifyListener) {
            textCrypto.removeModifyListener((ModifyListener) textCrypto.getData());
            textCrypto.setData(null);
        }
        textCrypto.setEditable(false);
        ti.setData(tcUnsavedEntry, null);
        
        toolbar.removeItemListener(Toolbar.ITEM_SAVE);
        toolbar.disableItems(Toolbar.ITEM_SAVE);
        mainmenu.removeItemListener(MainMenu.ITEM_SAVE);
        mainmenu.disableItems(MainMenu.ITEM_SAVE);

        toolbar.disableItems(GUI.T_STYLEELEMENTS);
        toolbar.removeItemListener(Toolbar.ITEM_BOLD);
        toolbar.removeItemListener(Toolbar.ITEM_ITALIC);
        toolbar.removeItemListener(Toolbar.ITEM_UNDERLINE);

        final ToolBar tbDiff = (ToolBar) ti.getData(tcToolbarDiff);
        tbDiff.setEnabled(true);
    }
    
    
    private boolean loadCryptoText(Crypto crypto, CryptoTab ct, Object userData) {
        final TextContent tc = new TextContent(parentComposite.getDisplay());
        
        CryptoEntry ce = ct.getCryptoEntry();
        CTabItem ti = tabs.get(ct);
        
        List listCrypto = getListCrypto(ti);
        StyledText textCrypto = getTextCrypto(ti);

        if(listCrypto == null || textCrypto == null)
        	return false;
        
        if(ti.getData(tcUnsavedEntry) != null) {
            MessageBox mb = new MessageBox(parentComposite.getShell(), SWT.YES | SWT.NO  | SWT.ICON_QUESTION);
            mb.setMessage(l10n._("You have unsaved data! Do you want discard your changes?"));
            if(mb.open() == SWT.NO) {
                listCrypto.select(0);
                textCrypto.setFocus();
                return false;
            }
        }

        /* enable edit mode only if first entry is selected */
        if(listCrypto.getSelectionIndex() == 0 && !ce.isReadonly())
            toolbar.enableItems(Toolbar.ITEM_EDIT);
        else
            toolbar.disableItems(Toolbar.ITEM_EDIT);

        disableTextItem(ti);
        
        int idx = listCrypto.getSelectionIndex();
        if(idx == -1)
            return false;
        
        if (CryptoData.class.isInstance(listCrypto.getData(tcCryptoData + Integer.toString(idx)))) {
            CryptoData cd = (CryptoData) listCrypto.getData(tcCryptoData + Integer.toString(idx));

            final Label labelChangedUID = (Label) ti.getData(tcLabelChangedUID);
            final Label labelChangedFPR = (Label) ti.getData(tcLabelChangedFPR);
            final Label labelChangedTimestamp = (Label) ti.getData(tcLabelChangedTimestamp);
            labelChangedUID.setText(GUI.escapeStr(gui.getUIDbyFPR(cd.getUKeyFPR())));
            labelChangedFPR.setText(cd.getUKeyFPR());
            labelChangedTimestamp.setText(cd.getDate().toString());
            
            try {
                textCrypto.setText("");
                String style = (String) ti.getData(tcDiffView); 
                String dt = Misc.b2s(crypto.decrypt(roleKeys.get(ce.getRKey_FPR()), cd.getEncryptedData().getBytes(), userData));
                if(style == null) {
	                if(dt == null) {
	                    textCrypto.setEnabled(false);
	                    toolbar.disableItems(Toolbar.ITEM_EDIT);
	                    toolbar.removeItemListener(Toolbar.ITEM_EDIT);                       
	                    errmsg.printErrorMessage(l10n._("Decryption of crypto entry failed."), EMessageTypes.TYPE_ERROR);

	                    return false;
	                } else {
	                    tc.setContent(dt);
	                    textCrypto.setText(tc.toPlainText());
	                    textCrypto.setStyleRanges(tc.getStyleRanges());
	                    textCrypto.setEnabled(true);
	                }
                }
                else {
                    toolbar.disableItems(Toolbar.ITEM_EDIT);
                    
                    String dt2 = "";
                    CryptoData cd2 = (CryptoData) listCrypto.getData(tcCryptoData + Integer.toString(idx + 1));
                    if(cd2 != null)
                    	dt2 = Misc.b2s(crypto.decrypt(roleKeys.get(ce.getRKey_FPR()), cd2.getEncryptedData().getBytes(), userData));

                    if(dt2 == null) {
	                    textCrypto.setEnabled(false);
	                    toolbar.disableItems(Toolbar.ITEM_EDIT);
	                    toolbar.removeItemListener(Toolbar.ITEM_EDIT);                       
	                    errmsg.printErrorMessage(l10n._("Decryption of crypto entry failed."), EMessageTypes.TYPE_ERROR);

	                    return false;
	                } else {
	                    tc.setContent(dt);
	                    String b[] = tc.toPlainText().split("\n");

	                    tc.setContent(dt2);
	                    String a[] = tc.toPlainText().split("\n");

	                    
	                    Diff d = new Diff(a, b);
	                    boolean reverse = style.equals("e");
	                    Diff.change script = d.diff_2(reverse);

	                    Base p;
	                    if(style.equals("e"))
	                        p = new EdPrint(a, b);
	                    else if(style.equals("c"))
	                        p = new ContextPrint(a, b);
		                else if(style.equals("u"))
	                        p = new UnifiedPrint(a, b);
		                else
	                        p = new NormalPrint(a, b);

	                    StringWriter wr = new StringWriter();
	                    
	                    p.setOutput(wr);
	                    p.print_header(
	                    		cd2 != null ? cd2.getDate() + " " + GUI.escapeStr(gui.getUIDbyFPR(cd2.getUKeyFPR())) : "- <empty>",
	                    		cd.getDate() + " " + GUI.escapeStr(gui.getUIDbyFPR(cd.getUKeyFPR()))
	                    );
	                    p.print_script(script);

	                    textCrypto.setText(wr.toString());
	                    textCrypto.setStyleRange(null);
	                    textCrypto.setEnabled(true);
	                }
                	
                }
            } catch (Exception excp) {
                textCrypto.setEnabled(false);
                toolbar.disableItems(Toolbar.ITEM_EDIT);
                toolbar.removeItemListener(Toolbar.ITEM_EDIT);                
                errmsg.printErrorMessage(l10n._("Decryption of crypto entry failed."), EMessageTypes.TYPE_ERROR);
                logger.debug(excp);

                return false;
            }
        }
        
        return true;
    }
    
    
    private void highlightText(StyledText styleText, final String search, final boolean enable) {
        final Color colorHighlight = parentComposite.getDisplay().getSystemColor(SWT.COLOR_YELLOW);
        
        if(enable == true) {
            if(search.length() < 2)
                return;                    
            
            final String searchText = styleText.getText();
            int pos = 0;
            int start = 0;                    
            while(pos != -1) {
                if((pos = searchText.toLowerCase().indexOf(search.toLowerCase(), start))  != -1) {
                    StyleRange sr = new StyleRange();
                    sr.start = pos;
                    sr.length = search.length();
                    sr.background = colorHighlight;
                    styleText.setStyleRange(sr);
                }
                start = pos + search.length();
            }
        } else {
            java.util.List<StyleRange> styleRanges = new ArrayList<StyleRange>();
            for(StyleRange sr :styleText.getStyleRanges()) {
                if(sr.background != null && (sr.background.getRed() == colorHighlight.getRed() &&
                        sr.background.getGreen() == colorHighlight.getGreen() &&
                        sr.background.getBlue() == colorHighlight.getBlue())) {
                    sr.background = bgColorTextCrypto;
                }

                styleRanges.add(sr);
            }
            styleText.setStyleRanges(styleRanges.toArray(new StyleRange[styleRanges.size()]));
        }
    }
    
    
    private void addSearchBar(final Composite comp, final StyledText styleText, final CTabItem ti) {
        final Composite fcomp = new Composite(comp, SWT.NONE);
        fcomp.setLayout(new GridLayout(5, false));
        final GridData gdcomp = new GridData(SWT.FILL, SWT.FILL, true, false);
        fcomp.setLayoutData(gdcomp);

        final Button btnCloseFind = new Button(fcomp, SWT.PUSH);
        btnCloseFind.setImage(GUI.getIcon(parentComposite.getDisplay(), "cancel.png"));
        btnCloseFind.setLayoutData(new GridData(SWT.LEFT, SWT.CENTER, false, false));
       
        Label l = new Label(fcomp, SWT.NONE);
        l.setText(l10n._("Find:"));
        l.setLayoutData(new GridData(SWT.LEFT, SWT.CENTER, false, false));
        
        final Text textFind = new Text(fcomp, SWT.SEARCH | SWT.CANCEL);
        textFind.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
        textFind.setEnabled(false);
        btnCloseFind.addListener(SWT.Selection, new Listener() {
            @Override
            public void handleEvent(Event event) {
                gdcomp.heightHint = 0;
                comp.layout();
                textFind.setEnabled(false);
                ti.setData(tcSearchBarVisible, null);
            }
        });            

        final Button btnNextFind = new Button(fcomp, SWT.PUSH);
        btnNextFind.setText(l10n._("Next"));
        btnNextFind.setImage(GUI.getIcon(parentComposite.getDisplay(), "forward.png"));
        btnNextFind.setLayoutData(new GridData(SWT.RIGHT, SWT.CENTER, false, false));
        btnNextFind.setEnabled(false);
        btnNextFind.addListener(SWT.Selection, new Listener() {
            
            @Override
            public void handleEvent(Event event) {
                if(styleText.getSelection() == null)
                    return;
                int start = styleText.getSelection().x + textFind.getText().length();
                final String searchText = styleText.getText();
                int pos = searchText.toLowerCase().indexOf(textFind.getText().toLowerCase(), start) == -1 ? 
                        searchText.toLowerCase().indexOf(textFind.getText().toLowerCase()) : 
                            searchText.toLowerCase().indexOf(textFind.getText().toLowerCase(), start);
                if(pos == -1)
                    return;
                styleText.setSelection(pos, pos + textFind.getText().length());                    
            }
        });
        
        final Button btnHighlightAll = new Button(fcomp, SWT.TOGGLE);
        btnHighlightAll.setText(l10n._("Highlight &all"));
        btnHighlightAll.setImage(GUI.getIcon(parentComposite.getDisplay(), "highlight.png"));
        btnHighlightAll.setLayoutData(new GridData(SWT.RIGHT, SWT.CENTER, false, false));
        btnHighlightAll.setEnabled(false);
        btnHighlightAll.setSelection(mainmenu.getItemSelection(MainMenu.ITEM_HIGHLIGHTALL));

        btnHighlightAll.addListener(SWT.Selection, new Listener() {
            @Override
            public void handleEvent(Event event) {
                highlightText(styleText, textFind.getText(), btnHighlightAll.getSelection());
            }
        });
        
        final Color bgColorTextItem = textFind.getBackground();
        final Color bgColorHighlight = new Color(Display.getDefault(), 255, 100, 100);            
        textFind.addModifyListener(new ModifyListener() {
            @Override
            public void modifyText(ModifyEvent event) {
                textFind.setBackground(bgColorTextItem);

                if(textFind.getText().length() < 2) {
                    btnNextFind.setEnabled(false);
                    btnHighlightAll.setEnabled(false);
                    highlightText(styleText, textFind.getText(), false);                    
                    return;
                }
                btnNextFind.setEnabled(true);
                btnHighlightAll.setEnabled(true);
                highlightText(styleText, textFind.getText(), false);
                highlightText(styleText, textFind.getText(), btnHighlightAll.getSelection());

                final String searchText = styleText.getText();
                int pos = searchText.toLowerCase().indexOf(textFind.getText().toLowerCase());
                if(pos == -1) {
                    styleText.setSelection(0);
                    textFind.setBackground(bgColorHighlight);
                    return;
                }
                styleText.setSelection(pos, pos + textFind.getText().length());
            }
        });
       
        textFind.addKeyListener(new KeyListener() {
            
            @Override
            public void keyReleased(KeyEvent event) { }
            
            @Override
            public void keyPressed(KeyEvent event) {
                if(event.keyCode == SWT.CR) {
                    if(styleText.getSelection() == null)
                        return;
                    int start = styleText.getSelection().x + textFind.getText().length();
                    final String searchText = styleText.getText();
                    int pos = searchText.toLowerCase().indexOf(textFind.getText().toLowerCase(), start) == -1 ? 
                            searchText.toLowerCase().indexOf(textFind.getText().toLowerCase()) : 
                                searchText.toLowerCase().indexOf(textFind.getText().toLowerCase(), start);
                    if(pos == -1)
                        return;
                    styleText.setSelection(pos, pos + textFind.getText().length());
                }
            }
        });
        
        mainmenu.removeItemListener(MainMenu.ITEM_FIND);
        mainmenu.addItemListener(MainMenu.ITEM_FIND, new Listener() {
            
            @Override
            public void handleEvent(Event event) {
                if(!textFind.getEnabled()) {
                    textFind.setEnabled(true);
                    textFind.setFocus();
                    gdcomp.heightHint = SWT.DEFAULT;
                    comp.layout();
                    ti.setData(tcSearchBarVisible, true);
                } else {
                    gdcomp.heightHint = 0;
                    comp.layout();
                    textFind.setEnabled(false);
                    ti.setData(tcSearchBarVisible, null);
                }
            }
        });
        gdcomp.heightHint = 0;        
    }
    
    
    public void addCryptoTab(final Customer cust, DBConn db, final Crypto crypto) {
        CTabItem tabCrypto = null;
        final CryptoTab ct = getCryptoTab(cust) != null ? getCryptoTab(cust) : new CryptoTab(cust);
        
      
        if(!tabs.containsKey(ct)) {
            Composite comp = new Composite(tabFolder, SWT.BORDER | SWT.FILL);
            comp.setLayout(new GridLayout(2, false));
            
            tabCrypto = new CTabItem(tabFolder, SWT.FILL);
            tabCrypto.setImage(GUI.getIcon(tabFolder.getDisplay(), "folder_grey.png"));

            String cname = GUI.escapeStr(GUI.shrinkStr(cust.getName(), 3, 10));
            tabCrypto.setText("[" + cname + "] (" +l10n._("New") + ")");
            tabCrypto.setData(tcUnsavedEntry, true);
            
            Label l = new Label(comp, SWT.NONE);
            l.setText(l10n._("New entry for ") + cust.getName());
            l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false, 2, 2));
            l.setFont(headline);
            
            l = new Label(comp, SWT.NONE);
            l.setText(l10n._("Subject:"));
            l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
            
            final Text textSubject = new Text(comp, SWT.BORDER);
            textSubject.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
            
            l = new Label(comp, SWT.NONE);
            l.setText(l10n._("Select a role:"));
            l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
            
            Map<Long, Role> roles;
            try {
                roles = db.getRoles();
            } catch (SRRoleNotFoundException e) {
                logger.debug(e);
    
                return;
            }
            
            final Combo cmbRoles = new Combo(comp, SWT.DROP_DOWN | SWT.BORDER | SWT.READ_ONLY);
            cmbRoles.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
            final Map<String, Role> roleMap = new HashMap<String, Role>();
            for(Iterator<UserRole> it = db.getUserRolesByFPR(crypto.getFPR(ECryptoPGP_KeyType.USER_KEY)).iterator(); it.hasNext(); ) {
                Role role = roles.get(it.next().getRoleID());
                if(!role.getReadonly()) {
                	roleMap.put(role.getName(), role);
                	cmbRoles.add(role.getName());
                }
            }
            if(cmbRoles.getItemCount() == 0) {
                errmsg.printErrorMessage(l10n._("You don't have any role to select."), EMessageTypes.TYPE_WARN);
                return;
            }
            else cmbRoles.select(0);
            
            l = new Label(comp, SWT.NONE);
            l.setText(l10n._("Description:"));
            l.setLayoutData(new GridData(SWT.BEGINNING, SWT.TOP, false, false));
            
            final Text textDescr = new Text(comp, SWT.BORDER | SWT.MULTI | SWT.V_SCROLL);
            GridData gd = new GridData(SWT.FILL, SWT.FILL, true, false);
            gd.minimumHeight = 100;
            gd.heightHint = 100;
            textDescr.setLayoutData(gd);
           
            Button saveButton = new Button(comp, SWT.BORDER);
            saveButton.setText(l10n._("Save"));
            saveButton.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
            Listener saveListener  = new Listener() {
                
                @Override
                public void handleEvent(Event event) {
                    try {
                        createCryptoEntry(ct, textSubject.getText(), textDescr.getText(), roleMap.get(cmbRoles.getText()));
                        if(tabs.get(ct).getData(tcUnsavedEntry) == null)
                            removeCTabItem(tabs.get(ct));
                    } catch (SRSigningKeyNotConfirmedException e) {
                        errmsg.printErrorMessage(l10n._("Signing key not confirmed!"), EMessageTypes.TYPE_WARN);
                        logger.debug(e);
                    } catch (SRSecretKeyNotFoundException e) {
                        logger.debug(e);
                    } catch (SRDBSQLException e) {
                        logger.debug(e);
                    }
                }
            };
            
            saveButton.addListener(SWT.Selection, saveListener);

            toolbar.addItemListener(Toolbar.ITEM_SAVE, saveListener);
            mainmenu.addItemListener(MainMenu.ITEM_SAVE, saveListener);
            
            Button cancelButton = new Button(comp, SWT.BORDER);
            cancelButton.setText(l10n._("Cancel"));
            cancelButton.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
    
            tabCrypto.setControl(comp);
            
            cancelButton.addListener(SWT.Selection, new Listener() {
                
                @Override
                public void handleEvent(Event event) {
                    removeCTabItem(tabs.get(ct));
                }
            });
    
            tabs.put(ct, tabCrypto);
        } else {
            tabCrypto = tabs.get(ct);
        }
        
        tabFolder.setSelection(tabCrypto);        
    }
    
    
    public void addCryptoTab(CryptoEntry ce, DBConn db, final Crypto crypto) {
        CTabItem tabCrypto = null;
        final CryptoTab ct = getCryptoTab(ce) != null ? getCryptoTab(ce) : new CryptoTab(ce);
        
        if(!tabs.containsKey(ct)) {
        	if(!roleKeys.containsKey(ce.getRKey_FPR())) {
        		String encKey = ce.getRole().getRoleKey().getEncryptedData();
        		
        		String roleKey;
				try {
					roleKey = Misc.b2s(crypto.decrypt(ECryptoPGP_KeyType.USER_KEY, encKey.getBytes(), PassphraseDialog.get(), null, tabFolder.getParent()));
				} catch (SRGenericException e) {
        			errmsg.printErrorMessage(l10n._("Can't decrypt the role key: ") + e.getLocalizedMessage(), EMessageTypes.TYPE_ERROR);
        			return;
				}
        		
        		roleKeys.put(ce.getRKey_FPR(), roleKey);
        	}

            final Composite comp = new Composite(tabFolder, SWT.BORDER | SWT.FILL);
            comp.setLayout(new GridLayout(2, false));
            
    		final Group gheader = new Group(comp, SWT.SHADOW_NONE);
            gheader.setText(l10n._("Details"));
    		gheader.setLayout(new GridLayout(2, true));
    		gheader.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));

    		Composite cl = new Composite(gheader, SWT.NONE);
    		cl.setLayout(new GridLayout(2, false));
    		cl.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
    		
    		Label l = new Label(cl, SWT.NONE);
    		l.setText(l10n._("Customer:"));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		l = new Label(cl, SWT.NONE);
    		l.setText(GUI.escapeStr(ce.getCustomer().getName()));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		
    		l = new Label(cl, SWT.NONE);
    		l.setText(l10n._("Entry:"));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		l = new Label(cl, SWT.NONE);
    		l.setText(GUI.escapeStr(ce.getSubject().trim()));
    		l.setToolTipText(ce.getDescription());
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		
    		l = new Label(cl, SWT.NONE);
    		l.setText(l10n._("Role:"));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		l = new Label(cl, SWT.NONE);
    		l.setText(GUI.escapeStr(ce.getRole().getName()));
    		l.setToolTipText(ce.getRole().getDescription());
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, true, false));

    		
    		l = new Label(cl, SWT.NONE);
    		l.setText(l10n._("Timestamp:"));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		final Label labelChangedTimestamp = new Label(cl, SWT.NONE);
    		labelChangedTimestamp.setText("");
    		labelChangedTimestamp.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, true, false));

    		

    		Composite cr = new Composite(gheader, SWT.NONE);
    		cr.setLayout(new GridLayout(2, false));
    		cr.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
    		
    		l = new Label(cr, SWT.NONE);
    		l.setText(l10n._("Created by UID:"));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		l = new Label(cr, SWT.NONE);
    		l.setText(GUI.escapeStr(gui.getUIDbyFPR(ce.getUKeyFPR())));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, true, false));

    		
    		l = new Label(cr, SWT.NONE);
    		l.setText(l10n._("Created by FPR:"));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		l = new Label(cr, SWT.NONE);
    		l.setText(ce.getUKeyFPR());
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, true, false));

    		
    		l = new Label(cr, SWT.NONE);
    		l.setText(l10n._("Changed by UID:"));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		final Label labelChangedUID = new Label(cr, SWT.NONE);
    		labelChangedUID.setText("");
    		labelChangedUID.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, true, false));

    		
    		l = new Label(cr, SWT.NONE);
    		l.setText(l10n._("Changed by FPR:"));
    		l.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));

    		final Label labelChangedFPR = new Label(cr, SWT.NONE);
    		labelChangedFPR.setText("");
    		labelChangedFPR.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, true, false));

    		
    		final Composite lc = new Composite(comp, SWT.NONE);
    		lc.setLayout(new GridLayout(2, false));
    		final GridData gdLc = new GridData(SWT.FILL, SWT.FILL, false, true, 1, 2);
    		lc.setLayoutData(gdLc);

    		final ToolBar tbDiff = new ToolBar(lc, SWT.BAR | SWT.FLAT);
    		final GridData gdDiff = new GridData(SWT.FILL, SWT.FILL, false, false, 1, 1);
    		tbDiff.setLayoutData(gdDiff);
    		
    		final ToolItem tiNoDiff = new ToolItem(tbDiff, SWT.RADIO);
    		tiNoDiff.setText(l10n._("content"));
    		tiNoDiff.setToolTipText(l10n._("show content"));
    		tiNoDiff.setSelection(true);

	        final ToolItem tiDiffNormal = new ToolItem(tbDiff, SWT.RADIO);
    		tiDiffNormal.setText(l10n._("diff"));
    		tiDiffNormal.setToolTipText(l10n._("show normal diff"));
    		tiDiffNormal.setData(tcDiffView, "n");
    		
	        final ToolItem tiDiffUnified = new ToolItem(tbDiff, SWT.RADIO);
    		tiDiffUnified.setText(l10n._("unified"));
    		tiDiffUnified.setToolTipText(l10n._("show unified diff"));
    		tiDiffUnified.setData(tcDiffView, "u");
    		
            final Sash sashList = new Sash(lc, SWT.VERTICAL);
            sashList.setLayoutData(new GridData(SWT.NONE, SWT.FILL, false, true, 1, 2));
            sashList.setBackgroundImage(GUI.getIcon(parentComposite.getDisplay(), "dots.png"));

            final List listCrypto = new List(lc, SWT.LEFT | SWT.BORDER | SWT.SINGLE | SWT.V_SCROLL);
            final GridData gdList = new GridData(SWT.FILL, SWT.FILL, false, true);
            listCrypto.setLayoutData(gdList);

            // Text content search bar & Text crypto
            final Composite rcomp = new Composite(comp, SWT.NONE);
            rcomp.setLayout(new GridLayout(1, false));
            rcomp.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
            
            sashList.addListener(SWT.Selection, new Listener() {
                
                @Override
                public void handleEvent(Event event) {
                    listCrypto.setVisible(true);
                    gdDiff.widthHint = event.x;
                    gdList.widthHint = event.x;
                    comp.layout();
                }
            });
            
            // If the collapse menu item is checked.
            if(mainmenu.getItemSelection(MainMenu.ITEM_COLLAPSELIST)) {
                gdDiff.widthHint = 0;
                gdList.widthHint = 0;
                listCrypto.setVisible(false);
                comp.layout();
            }
            
            final StyledText textCrypto = new StyledText(rcomp, SWT.MULTI | SWT.BORDER | SWT.V_SCROLL | SWT.H_SCROLL);
            textCrypto.setFont(monospace);            
            textCrypto.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
            textCrypto.setBackground(bgColorTextCrypto);

            // Event handler for styled text.
            textCrypto.addExtendedModifyListener(new ExtendedModifyListener() {
                public void modifyText(ExtendedModifyEvent e) {
                  handleExtendedModify(e, textCrypto);
                }
              });
            
            textCrypto.addSelectionListener(new SelectionListener() {
                
                @Override
                public void widgetSelected(SelectionEvent event) {
                    int length = Math.abs(event.x - event.y);
                    int start = event.x < event.y ? event.x : event.y;
                    StyleRange [] styles = textCrypto.getStyleRanges(start, length);
                    if(styles.length == 0) {
                        toolbar.setBoldItemChecked(false);
                        toolbar.setItalicItemChecked(false);
                        toolbar.setUnderlineItemChecked(false);
                        toolbar.setColorItemChecked(false);
                    }
                    for(StyleRange style : styles) {
                        toolbar.setBoldItemChecked(((style.fontStyle & SWT.BOLD) == SWT.BOLD));
                        toolbar.setItalicItemChecked((style.fontStyle & SWT.ITALIC) == SWT.ITALIC);
                        toolbar.setUnderlineItemChecked((style.fontStyle & SWT.SINGLE) == SWT.SINGLE);
                    }
                }
                
                @Override
                public void widgetDefaultSelected(SelectionEvent event) {
                    // TODO Auto-generated method stub
                }
            });
            
            textCrypto.addListener(SWT.MenuDetect, new Listener() {
                public void handleEvent(Event event) {
                    showCryptoTextContextMenu((Shell) tabFolder.getParent(), tabFolder.getDisplay(), ct, textCrypto, event);
                }                
            });
            
            final CTabItem _tabCrypto = new CTabItem(tabFolder, SWT.FILL | SWT.CLOSE);
            String cname = GUI.escapeStr(GUI.shrinkStr(ce.getCustomer().getName(), 3, 10));
            _tabCrypto.setText("[" + cname + "] " + GUI.escapeStr(ce.getSubject()).trim());
            _tabCrypto.setControl(comp);
            _tabCrypto.setImage(GUI.getIcon(tabFolder.getDisplay(), ce.isReadonly() ? "folder_yellow.png" : "folder_green_open.png"));

            // Setup the data fields for the crypto tab.
            _tabCrypto.setData(tcTextKey, textCrypto);
            _tabCrypto.setData(tcVersionListKey, listCrypto);
            _tabCrypto.setData(tcLabelChangedUID, labelChangedUID);
            _tabCrypto.setData(tcLabelChangedFPR, labelChangedFPR);
            _tabCrypto.setData(tcLabelChangedTimestamp, labelChangedTimestamp);
            _tabCrypto.setData(tcToolbarDiff, tbDiff);
            
            // Disable edit mode per default.
            disableTextItem(tabCrypto);
           
            tabs.put(ct, _tabCrypto);

            listCrypto.addListener(SWT.Selection, new Listener() {
				public void handleEvent(Event e) {
				    if(!loadCryptoText(crypto, ct, tabFolder.getParent())) {
				        removeCTabItem(_tabCrypto);
				        return;
				    }
				}
			});


    		final SelectionListener selDiff = new SelectionListener() {
				@Override
				public void widgetSelected(SelectionEvent arg0) {
					tabs.get(ct).setData(tcDiffView, arg0.widget.getData(tcDiffView));
				    if(!loadCryptoText(crypto, ct, tabFolder.getParent())) {
				        removeCTabItem(_tabCrypto);
				        return;
				    }
				}
				
				@Override
				public void widgetDefaultSelected(SelectionEvent arg0) {
				}
			};
    		tiNoDiff.addSelectionListener(selDiff);
    		tiDiffNormal.addSelectionListener(selDiff);
    		tiDiffUnified.addSelectionListener(selDiff);

    		// Initialize search bar
            addSearchBar(rcomp, textCrypto, _tabCrypto);

            // Fill the list.
            if(!refreshListCryptoItem(ct)) {
                removeCTabItem(_tabCrypto);
                return;
            }
            
            tabCrypto = _tabCrypto;
        } else {
            tabCrypto = tabs.get(ct);
        }
        
        tabFolder.setSelection(tabCrypto);
    }

    
    void handleExtendedModify(ExtendedModifyEvent event, StyledText textCrypto) {
        ToolItem boldItem      = toolbar.getBoldItem();
        ToolItem italicItem    = toolbar.getItalicItem();
        ToolItem underlineItem = toolbar.getUnderlineItem();

    	if (event.length == 0)
          return;
        StyleRange style;
        if (event.length == 1
            || textCrypto.getTextRange(event.start, event.length).equals(
                textCrypto.getLineDelimiter())) {
          int caretOffset = textCrypto.getCaretOffset();
          style = null;
          if (caretOffset < textCrypto.getCharCount())
            style = textCrypto.getStyleRangeAtOffset(caretOffset);
          if (style != null) {
            style = (StyleRange) style.clone();
            style.start = event.start;
            style.length = event.length;
          } else {
            style = new StyleRange(event.start, event.length, null, null,
                SWT.NORMAL);
          }
          if (boldItem.getSelection())
            style.fontStyle |= SWT.BOLD;
          if (italicItem.getSelection())
            style.fontStyle |= SWT.ITALIC;
          if (underlineItem.getSelection())
              style.fontStyle |= SWT.SINGLE;
          if (!style.isUnstyled())
        	  textCrypto.setStyleRange(style);
        } else {
          // paste occurring, have text take on the styles it had when it was
          // cut/copied
          for (int i = 0; i < cachedStyles.size(); i++) {
            style = (StyleRange) cachedStyles.elementAt(i);
            StyleRange newStyle = (StyleRange) style.clone();
            newStyle.start = style.start + event.start;
            textCrypto.setStyleRange(newStyle);
          }
        }

        //logger.debug("CryptoData: " + textCrypto.getCaretOffset());      
    }
    

    private boolean setStyle(StyledText sText, int tbStyleBtn) {
        Color red = parentComposite.getDisplay().getSystemColor(SWT.COLOR_RED);
        Color green = parentComposite.getDisplay().getSystemColor(SWT.COLOR_GREEN);
        Color blue = parentComposite.getDisplay().getSystemColor(SWT.COLOR_BLUE);
        Color black = parentComposite.getDisplay().getSystemColor(SWT.COLOR_BLACK);
        
        Point sel = null;
        sel = sText.getSelectionRange();
        
        if ((sel == null) || (sel.y == 0))
            return false;

        StyleRange style;
        for (int i = sel.x; i < sel.x + sel.y; i++) {
            StyleRange range = sText.getStyleRangeAtOffset(i);
            if (range != null) {
                style = (StyleRange) range.clone();
                style.start = i;
                style.length = 1;
            } else {
                style = new StyleRange(i, 1, null, null, SWT.NORMAL);
            }
            if (tbStyleBtn == Toolbar.ITEM_BOLD) {
                style.fontStyle ^= SWT.BOLD;
            } else if (tbStyleBtn == Toolbar.ITEM_ITALIC) {
                style.fontStyle ^= SWT.ITALIC;
            } else if (tbStyleBtn == Toolbar.ITEM_UNDERLINE) {
                style.fontStyle ^= SWT.SINGLE;
            } else if (tbStyleBtn == Toolbar.ITEM_BLACK) {
                style.foreground = black;
            } else if (tbStyleBtn == Toolbar.ITEM_RED) {
                style.foreground = red;
            } else if (tbStyleBtn == Toolbar.ITEM_BLUE) {
                style.foreground = blue;
            } else if (tbStyleBtn == Toolbar.ITEM_GREEN) {
                style.foreground = green;
            }
            sText.setStyleRange(style);
        }
        sText.setSelectionRange(sel.x + sel.y, 0);
        
        return true;
    }
    
    
    protected StyledText createStyledText(Composite parent, int style) {
        return new StyledText(parent, style);
    }
}
