package de.ibh.StrongRoom.DB;

import java.sql.Timestamp;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.ibh.StrongRoom.Crypto.CryptoBackend.ECryptoPGP_KeyType;


public class CryptoData implements ISignedRow, IEncryptedRow {
	private CryptoEntry ce;
	private Timestamp date;
	private String enc_data;
	private String ukey_fpr;
	private String sign_rkey;
	private String sign_ukey;
	
	public CryptoData(final CryptoEntry ce, final Timestamp tstamp, final String enc_data, final String ukey_fpr, final String sign_rkey, final String sign_ukey) {
		this.ce = ce;
		this.date = tstamp != null ? tstamp : new java.sql.Timestamp(new java.util.Date().getTime());
		this.enc_data = enc_data;
		this.ukey_fpr = ukey_fpr;
		this.sign_rkey = sign_rkey;
		this.sign_ukey = sign_ukey;
	}

	public CryptoEntry getCE() {
		return ce;
	}

	public Timestamp getDate() {
		return date;
	}

	public String getEncryptedData() {
		return this.enc_data;
	}

	public String getUKeyFPR() {
		return this.ukey_fpr;
	}

	@Override
	public Map<ECryptoPGP_KeyType, String> getSignKeyMap() {
		Map<ECryptoPGP_KeyType, String> m = new HashMap<ECryptoPGP_KeyType, String>();

		m.put(ECryptoPGP_KeyType.ROLE_KEY, sign_rkey);
		m.put(ECryptoPGP_KeyType.USER_KEY, sign_ukey);

		return m;
	}

	@Override
	public String[] getSignStrings() {
		return new String[] {
				Long.toString(ce.getId()),
				date.toString(),
				enc_data,
				ukey_fpr,
		};
	}

	@Override
	public Set<ECryptoPGP_KeyType> getEncKeySet() {
		Set<ECryptoPGP_KeyType> ek = new HashSet<ECryptoPGP_KeyType>();

		ek.add(ECryptoPGP_KeyType.MASTER_KEY);
		ek.add(ECryptoPGP_KeyType.ROLE_KEY);

		return ek;
	}

	@Override
	public void setEncryptedData(String enc_data) {
		this.enc_data = enc_data;
	}

	@Override
	public void setSigningData(ECryptoPGP_KeyType kt, String sig) {
		if(kt == ECryptoPGP_KeyType.ROLE_KEY)
			sign_rkey = sig;
		else if(kt == ECryptoPGP_KeyType.USER_KEY)
			sign_ukey = sig;
		else
			throw new IllegalArgumentException();
	}
}
