//---------------------------------------------------------------------------
// Freely available from Media Art Online (http://www.media-art-online.org/).
// Copyright (C) 2006 Media Art Online (cafe@media-art-online.org)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// PROJECT: PGP/GnuPG Applications
// MODULE:  GnuPG support
// FILE:    GnuPGSystemKey.java
//
// DESCRIPTION:
//
// REVISION HISTORY:
//  Name      Date        Description
//  ----      ----------  -----------
//  ks91      2006/10/18  initial version
//
// $Id: //depot/scm/main/java/org/media_art_online/gnupg/GnuPGSystemKey.java#1 $
// $DateTime: 2006/11/15 13:44:06 $
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Package
//---------------------------------------------------------------------------
package org.media_art_online.gnupg;

//---------------------------------------------------------------------------
// Import
//---------------------------------------------------------------------------
import java.io.*;
import java.net.*;
import java.util.Iterator;

public class GnuPGSystemKey {

	public GnuPGSystemKey(String sFingerprint, URL url) {
		_sFingerprint  = sFingerprint;
		_url = url;
	}

	public File getFile() {

		FileOutputStream output = null;
		InputStream      input  = null;
		File             file   = null;

		try {
			file = File.createTempFile(GnuPG.S_GNUPG, GnuPG.S_EXT_GPG);

			output = new FileOutputStream(file);
			input = _url.openStream();

			int c;

			while ((c = input.read()) > 0) {
				output.write(c);
			}

			output.close();
			output = null;

		} catch (Throwable ex) {
			System.out.println(ex.toString());

		} finally {

			try {

				if (output != null) {
					output.close();
				}
				if (input != null) {
					input.close();
				}

			} catch (Exception unused) {
			}
		}

		return (file);
	}

	public String getFingerprint() {
		return (_sFingerprint);
	}

	public URL getURL() {
		return (_url);
	}

	boolean isImported() {

		boolean isFound = false;

		for (Iterator<GnuPGUser> iterator = GnuPG.listPublicKeys().iterator();
		 iterator.hasNext(); ) {

			GnuPGUser user = iterator.next();

			if (user.getPrimaryKey().getFingerprint().equals(_sFingerprint)) {
				isFound = true;
				break;
			}
		}

		return (isFound);
	}

//===========================================================================
// BARRIER: Anything below is not open to other files.
//===========================================================================

	private String _sFingerprint;

	private URL _url;
};

// GnuPGSystemKey.java
